from __future__ import annotations
import re
import socket
from hedra.distributed.discovery.dns.core.record import RecordType
from typing import Tuple, Union, Optional
from urllib.parse import urlparse
from .exceptions import (
    InvalidHost,
    InvalidIP
)


ip_pattern = '(?P<host>[^:/ ]+).?(?P<port>[0-9]*).*'
match_pattern = re.compile(ip_pattern)



class URL:



    def __init__(
        self, 
        url: str,
        port: Optional[int]=None
    ):

        self._default_ports = {
            'tcp': 53,
            'udp': 53,
            'tcps': 853,
            'http': 80,
            'https': 443,
        }
        

        self.url = url
        self.parsed = urlparse(url)

        self.host = self.parsed.hostname

        if port is None:
            port = self.parsed.port

        self.port = port

        if self.host is None:
            (
                _,
                host,
                _
            ) = self.parse_netloc()

            self.host = host

        self.is_ssl = False
        if self.parsed.scheme in ['tcps', 'https', 'msyncs']:
            self.is_ssl = True

        self.ip_type = self.get_ip_type(
            self.host
        )

        if self.ip_type is None:

            matches = re.search(
                ip_pattern,
                self.url
            )
            self.host = matches.group('host')
            self.port = matches.group('port')

            if self.port:
                self.port = int(self.port)
        

        if self.port is None or self.port == '':
            self.port = self._default_ports.get(
                self.parsed.scheme,
                80
            )

        self.domain_protocol_map = {
            "tcp": "tcp",
            "udp": "udp",
            "tcps": "tcp",
            "http": "tcp",
            "https": "tcp"
        }

        self.address = (
            self.host, 
            self.port
        )

        self.is_msync = self.parsed.scheme in ['msync', 'msyncs']

    def __str__(self):
        return self.url

    def __eq__(self, other):
        return str(self) == str(other)

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self))

    def copy(self):
        return URL(self.url)
    
    def parse_netloc(self):

        authentication: Union[str, None] = None
        port: Union[str, None] = None

        host = self.parsed.netloc

        if '@' in host:
            authentication, host = host.split('@')

        if ':' in host:
            host, port = host.split(':')

        if port:
            port = int(port)

        return (
            authentication,
            host,
            port
        )
            

    def to_ptr(self):
        if self.ip_type is RecordType.A:
            reversed_hostname =  '.'.join(
                self.parsed.hostname.split('.')[::-1]
            )

            return f'{reversed_hostname}.in-addr.arpa'
        
        raise InvalidIP(self.parsed.hostname)
    
    def get_ip_type(
        self,
        hostname: str
    ):

        if ':' in hostname:
            # ipv6
            try:
                socket.inet_pton(socket.AF_INET6, hostname)
            except OSError:
                raise InvalidHost(hostname)
            
            return RecordType.AAAA
        
        try:
            socket.inet_pton(socket.AF_INET, hostname)
        except OSError:
            # domain name
            pass
        else:
            return RecordType.A

    
    @property
    def domain_protocol(self):
        return self.domain_protocol_map.get(self.parsed.scheme, "udp")



from enum import Enum


class PingType(Enum):
    ICMP='ICMP'
    TCP='TCP'
    UDP='UDP'


class PingTypesMap:
    
    def __init__(self) -> None:
        self._ping_types = {
            'icmp': PingType.ICMP,
            'tcp': PingType.TCP,
            'udp': PingType.UDP
        }

        self._ping_type_name = {
            PingType.ICMP: 'icmp',
            PingType.TCP: 'tcp',
            PingType.UDP: 'udp'
        }

    def get(self, ping_type_name: str) -> PingType:
        return self._ping_types.get(ping_type_name)
    
    def get_name(self, ping_type: PingType) -> str:
        return self._ping_type_name.get(ping_type)


from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class CosmosDBConfig(BaseModel):
    account_uri: str
    account_key: str
    database: str
    events_container: str='events'
    metrics_container: str='metrics'
    streams_container: str='streams'
    experiments_container: str='experiments'
    system_metrics_container: str='system_metrics'
    events_partition_key: str='name'
    metrics_partition_key: str='name'
    streams_partition_key: str='name'
    experiments_partition_key: str='experiment_name'
    variants_partition_key: str='variant_name'
    mutations_partition_key: str='mutation_name'
    system_metrics_partition: str='name'
    analytics_ttl: int=0
    reporter_type: ReporterTypes=ReporterTypes.CosmosDB


import numpy
from typing import Union, List, Dict
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.task.hook import TaskHook
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.versioning.flags.types.unstable.flag import unstable


@unstable
class Stream:

    __slots__ = (
        'last_completed',
        'last_batch_size',
        'action',
        'completed'
    )

    def __init__(self) -> None:
        self.last_completed = 0
        self.last_batch_size = 0
        self.action = None
        self.completed: List[BaseResult] = []

    @property
    def completed_count(self):
        return len(self.completed)

    @property
    def succeeded(self) -> int:
        return len([
            result for result in self.completed if result.error is None
        ])
    
    @property
    def failed(self) -> int:
        succeeded = self.succeeded
        return len(self.completed) - succeeded
    
    @property
    def timings(self) -> List[Dict[str, float]]:

        timings = {
            'total': [],
            'waiting': [],
            'connecting': [],
            'writing': [],
            'reading': []
        }

        for result in self.completed:
            timings['total'].append(result.complete - result.start)
            timings['waiting'].append(result.start - result.wait_start)
            timings['connecting'].append(result.connect_end - result.start)
            timings['writing'].append(result.write_end - result.connect_end)
            timings['reading'].append(result.complete - result.write_end)

        stream_timings = {}

        for timing_group_name, timing_group in timings.items():
            if len(timing_group) > 0:
                stream_timings[timing_group_name] = {
                    'maximum': max(timing_group),
                    'minimum': min(timing_group),
                    'median': numpy.median(timing_group),
                    'mean': numpy.mean(timing_group),
                    'stdev': numpy.std(timing_group),
                    'variance': numpy.var(timing_group)
                }

            else:
                stream_timings[timing_group_name] = {
                    'maximum': 0,
                    'minimum': 0,
                    'median': 0,
                    'mean': 0,
                    'stdev': 0,
                    'variance': 0
                }



        return stream_timings


    async def execute_action(self, hook: Union[ActionHook, TaskHook]):
        try:
            result: BaseResult = await hook.session.execute_prepared_request(
                hook.action
            )
        except RuntimeError as runtime_error:
            result = runtime_error
        
        self.completed.append(result)

        return result


from .base_event import BaseEvent


class InformationalResponseReceived(BaseEvent):
    event_type='INFORMATIONAL_RESPONSE_RECEIVED'
    __slots__ = (
        'stream_id',
        'headers',
        'priority_updated'
    )
    """
    The InformationalResponseReceived event is fired when an informational
    response (that is, one whose status code is a 1XX code) is received from
    the remote peer.

    The remote peer may send any number of these, from zero upwards. These
    responses are most commonly sent in response to requests that have the
    ``expect: 100-continue`` header field present. Most users can safely
    ignore this event unless you are intending to use the
    ``expect: 100-continue`` flow, or are for any reason expecting a different
    1XX status code.

    .. versionadded:: 2.2.0

    .. versionchanged:: 2.3.0
       Changed the type of ``headers`` to :class:`HeaderTuple
       <hpack:hpack.HeaderTuple>`. This has no effect on current users.

    .. versionchanged:: 2.4.0
       Added ``priority_updated`` property.
    """
    def __init__(self):
        #: The Stream ID for the stream this informational response was made
        #: on.
        self.stream_id = None

        #: The headers for this informational response.
        self.headers = None

        #: If this response also had associated priority information, the
        #: associated :class:`PriorityUpdated <h2.events.PriorityUpdated>`
        #: event will be available here.
        #:
        #: .. versionadded:: 2.4.0
        self.priority_updated = None

    def __repr__(self):
        return "<InformationalResponseReceived stream_id:%s, headers:%s>" % (
            self.stream_id, self.headers
        )


import uuid
import json
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.playwright import (
    PlaywrightCommand,
    Page,
    Input,
    URL,
    Options,
    MercuryPlaywrightClient,
    ContextConfig
)
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.engines.types.common.types import RequestTypes
from hedra.data.parsers.parser_types.common.base_parser import BaseParser
from hedra.data.parsers.parser_types.common.parsing import (
    normalize_headers,
    parse_tags
)
from typing import Any, Coroutine, Dict
from .playwright_action_validator import (
    PlaywrightActionValidator,
    PlaywrightInputValidator,
    PlaywrightOptionsValidator,
    PlaywrightPageValidator,
    PlaywrightURLValidator
)


class PlaywrightActionParser(BaseParser):

    def __init__(
        self,
        config: Config,
        options: Dict[str, Any]={}
    ) -> None:
        super().__init__(
            PlaywrightActionParser.__name__,
            config,
            RequestTypes.PLAYWRIGHT,
            options
        )

    async def parse(
        self, 
        action_data: Dict[str, Any],
        stage: str
    ) -> Coroutine[Any, Any, Coroutine[Any, Any, ActionHook]]:
        
        normalized_headers = normalize_headers(action_data)
        tags_data = parse_tags(action_data)

        playwright_input_args = action_data.get('args')
        if isinstance(playwright_input_args, (str, bytes, bytearray,)):
            playwright_input_args = playwright_input_args.split(',')

        playwright_options_extra = action_data.get('extra')
        if isinstance(playwright_options_extra, (str, bytes, bytearray,)):
            playwright_options_extra = json.loads(playwright_options_extra)

        generator_action = PlaywrightActionValidator(
            name=action_data.get('name'),
            command=action_data.get('command'),
            page=PlaywrightPageValidator(
                selector=action_data.get('selector'),
                attribute=action_data.get('attribute'),
                x_coordinate=action_data.get('x_coordinate'),
                y_coordinate=action_data.get('y_coordinate'),
                frame=action_data.get('frame')
            ),
            url=PlaywrightURLValidator(
                location=action_data.get('location'),
                headers=normalized_headers
            ),
            input=PlaywrightInputValidator(
                key=action_data.get('key'),
                text=action_data.get('text'),
                expression=action_data.get('expression'),
                args=playwright_input_args,
                filepath=action_data.get('filepath'),
                file=action_data.get('file'),
                path=action_data.get('path'),
                option=action_data.get('option'),
                by_label=action_data.get('by_label', False),
                by_value=action_data.get('by_value', False)
            ),
            options=PlaywrightOptionsValidator(
                event=action_data.get('event'),
                option=action_data.get('option'),
                is_checked=action_data.get('is_checked', False),
                timeout=action_data.get('timeout'),
                extra=playwright_options_extra,
                switch_by=action_data.get('switch_by')
            ),
            weight=action_data.get('weight'),
            order=action_data.get('order'),
            user=action_data.get('user'),
            tags=tags_data
        )

        action = PlaywrightCommand(
            generator_action.name,
            generator_action.command,
            url=URL(**generator_action.url.dict(
                exclude_none=True
            )),
            page=Page(**generator_action.page.dict(
                exclude_none=True
            )),
            input=Input(**generator_action.input.dict(
                exclude_none=True
            )),
            options=Options(**generator_action.options.dict(
                exclude_none=True
            )),
            user=generator_action.user,
            tags=[
                tag.dict() for tag in generator_action.tags
            ]
        )

        session = MercuryPlaywrightClient(
            concurrency=self.config.batch_size,
            timeouts=self.timeouts,
            group_size=self.config.group_size
        )

        await session.setup(
            config=ContextConfig(
                browser_type=self.config.browser_type,
                device_type=self.config.device_type,
                locale=self.config.locale,
                geolocation=self.config.geolocation,
                permissions=self.config.permissions,
                color_scheme=self.config.color_scheme,
                options=self.config.playwright_options
            )
        )

        await session.prepare(action)

        hook = ActionHook(
            f'{stage}.{generator_action.name}',
            generator_action.name,
            None,
            sourcefile=generator_action.sourcefile,
        )

        hook.session = session
        hook.action = action
        hook.stage = stage
        hook.context = SimpleContext()
        hook.hook_id = uuid.uuid4()
      
        hook.metadata.order = generator_action.order
        hook.metadata.weight = generator_action.weight
        hook.metadata.tags = generator_action.tags
        hook.metadata.user = generator_action.user

        return hook






from __future__ import annotations
import asyncio
import inspect
import traceback
from collections import defaultdict
from typing import Dict, List, Any, Union
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.registrar import registrar
from hedra.core.hooks.types.task.hook import TaskHook
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.graphs.transitions.common.base_edge import BaseEdge
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.optimize.optimize import Optimize
from hedra.core.graphs.stages.execute.execute import Execute
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from hedra.reporting.system.system_metrics_set_types import MonitorGroup


ExecuteHooks = List[Union[ActionHook , TaskHook]]


class OptimizeEdge(BaseEdge[Optimize]):

    def __init__(self, source: Optimize, destination: Stage) -> None:
        super(
            OptimizeEdge,
            self
        ).__init__(
            source,
            destination
        )

        self.requires = [
            'setup_stage_experiment_config',
            'execute_stage_streamed_analytics',
            'setup_stage_configs',
            'execute_stage_setup_hooks',
            'execute_stage_setup_config',
            'execute_stage_setup_by',
            'setup_stage_ready_stages',
            'setup_stage_candidates',
            'execute_stage_results',
            'session_stage_monitors',
        ]

        self.provides = [
            'setup_stage_experiment_config',
            'execute_stage_streamed_analytics',
            'setup_stage_configs',
            'optimize_stage_optimized_params',
            'optimize_stage_optimized_configs',
            'optimize_stage_optimized_hooks',
            'execute_stage_setup_by',
            'execute_stage_results',
            'optimize_stage_monitors',
            'session_stage_monitors'
        ]

        self.assigned_candidates = []
        self.edge_data = {}

    async def transition(self):

        try:
            selected_optimization_candidates = self.generate_optimization_candidates()

            self.edge_data['optimize_stage_candidates'] = selected_optimization_candidates
            
            self.source.context.update(self.edge_data)
            
            for event in self.source.dispatcher.events_by_name.values():
                self.source.context.update(self.edge_data)

                event.source.stage_instance = self.source
                event.context.update(self.edge_data)
                
                if event.source.context:
                    event.source.context.update(self.edge_data)

            if len(selected_optimization_candidates) > 0:
                self.source.generation_optimization_candidates = len(selected_optimization_candidates)

                if self.timeout and self.skip_stage is False:
                    await asyncio.wait_for(self.source.run(), timeout=self.timeout)

                elif self.skip_stage is False:
                    await self.source.run()

            for provided in self.provides:
                self.edge_data[provided] = self.source.context[provided]

            if self.destination.context is None:
                self.destination.context = SimpleContext()
            
            self._update(self.destination)
            if len(selected_optimization_candidates) > 0:

                for optimization_candidate in selected_optimization_candidates.values():

                    if optimization_candidate.context is None:
                        optimization_candidate.context = SimpleContext()

                    self._update(optimization_candidate)

                    optimization_candidate.state = StageStates.OPTIMIZED

            self.visited.append(self.source.name)

        except Exception as edge_exception:
            self.exception = edge_exception

        return None, self.destination.stage_type


    def _update(self, destination: Stage):

        for edge_name in self.history:

            history = self.history[edge_name]

            if self.next_history.get(edge_name) is None:
                self.next_history[edge_name] = {}

            self.next_history[edge_name].update(history)

        if self.skip_stage is False:

            if self.next_history.get((self.source.name, destination.name)) is None:
                self.next_history[(self.source.name, destination.name)] = {}

            optimized_config: Stage = self.edge_data['optimize_stage_optimized_configs'].get(destination.name)
            optimzied_hooks: Stage = self.edge_data['optimize_stage_optimized_hooks'].get(destination.name)
            stage_setup_by: str = self.edge_data['execute_stage_setup_by'].get(destination.name)

            session_stage_monitors: MonitorGroup = self.edge_data['session_stage_monitors']
            session_stage_monitors.update(
                self.edge_data['optimize_stage_monitors']
            )

            self.next_history[(self.source.name, destination.name)].update({
                'setup_stage_configs': self.edge_data['optimize_stage_optimized_configs'],
                'optimize_stage_optimized_params': self.edge_data['optimize_stage_optimized_params'],
                'setup_stage_candidates': list(self.edge_data['optimize_stage_candidates'].keys()),
                'setup_stage_ready_stages': self.edge_data['setup_stage_ready_stages'],
                'execute_stage_setup_config': optimized_config,
                'execute_stage_setup_hooks': optimzied_hooks,
                'execute_stage_setup_by': stage_setup_by,
                'session_stage_monitors': session_stage_monitors
            }) 

    def split(self, edges: List[OptimizeEdge]) -> None:

        optimize_stage_config: Dict[str, Any] = self.source.to_copy_dict()

        optimize_stage_copy: Optimize = type(self.source.name, (Optimize, ), self.source.__dict__)()
        
        for copied_attribute_name, copied_attribute_value in optimize_stage_config.items():
            if inspect.ismethod(copied_attribute_value) is False:
                setattr(
                    optimize_stage_copy, 
                    copied_attribute_name, 
                    copied_attribute_value
                )

        user_hooks: Dict[str, Dict[str, Hook]] = defaultdict(dict)
        for hooks in registrar.all.values():
            for hook in hooks:
                if hasattr(self.source, hook.shortname) and not hasattr(Execute, hook.shortname):
                    user_hooks[self.source.name][hook.shortname] = hook._call

        optimize_stage_copy.dispatcher = self.source.dispatcher.copy()

        for event in optimize_stage_copy.dispatcher.events_by_name.values():
            event.source.stage_instance = optimize_stage_copy

        minimum_edge_idx = min([edge.transition_idx for edge in edges])
        
        optimize_stage_copy.context = SimpleContext()
        optimize_stage_copy.context.update(self.source.context)

        for event in optimize_stage_copy.dispatcher.events_by_name.values():
            event.source.stage_instance = optimize_stage_copy 
            event.source.stage_instance = optimize_stage_copy
            event.source.stage_instance.context = optimize_stage_copy.context
            event.source.context = optimize_stage_copy.context

            if event.source.shortname in user_hooks[optimize_stage_copy.name]:
                hook_call = user_hooks[optimize_stage_copy.name].get(event.source.shortname)

                hook_call = hook_call.__get__(optimize_stage_copy, optimize_stage_copy.__class__)
                setattr(optimize_stage_copy, event.source.shortname, hook_call)

                event.source._call = hook_call

            else:            
                event.source._call = getattr(optimize_stage_copy, event.source.shortname)
                event.source._call = event.source._call.__get__(optimize_stage_copy, optimize_stage_copy.__class__)
                setattr(optimize_stage_copy, event.source.shortname, event.source._call)

        self.source = optimize_stage_copy

        if minimum_edge_idx < self.transition_idx:
            self.skip_stage = True

    def generate_optimization_candidates(self) -> Dict[str, Stage]:
        
        execute_stages: Dict[str, Execute] = self.stages_by_type.get(StageTypes.EXECUTE)
        optimize_stages = self.stages_by_type.get(StageTypes.OPTIMIZE).items()
        setup_stages = self.stages_by_type.get(StageTypes.SETUP).items()
        path_lengths: Dict[str, int] = self.path_lengths.get(self.source.name)

        all_paths = self.all_paths.get(self.source.name, [])

        optimization_candidates: Dict[str, Stage] = {}

        for stage_name, stage in execute_stages.items():
            if stage_name in all_paths:
                stage.context['execute_stage_setup_config'] = self.edge_data['execute_stage_setup_config']
                stage.context['execute_stage_setup_by'] = self.edge_data['execute_stage_setup_by']
                optimization_candidates[stage_name] = stage

        selected_optimization_candidates: Dict[str, Stage] = {}
        following_optimize_stage_distances = [
            path_length for stage_name, path_length in path_lengths.items() if stage_name in optimize_stages
        ]

        for stage_name in path_lengths.keys():
            stage_distance = path_lengths.get(stage_name)

            if stage_name in optimization_candidates:

                if len(following_optimize_stage_distances) > 0 and stage_distance <= min(following_optimize_stage_distances):
                    selected_optimization_candidates[stage_name] = optimization_candidates.get(stage_name)

                elif len(following_optimize_stage_distances) == 0:
                    selected_optimization_candidates[stage_name] = optimization_candidates.get(stage_name)
        
        return {
            candidate: candidate_config for candidate, candidate_config in selected_optimization_candidates.items() if candidate in self.edge_data['setup_stage_configs']
        }
    
    def setup(self) -> None:

        max_batch_size = 0
        execute_stage_setup_config: Config = None
        execute_stage_setup_hooks: Dict[str, ExecuteHooks] = {}
        execute_stage_setup_by: str = None
        execute_stage_streamed_analytics: Dict[str, List[StreamAnalytics]]= defaultdict(list)
        setup_stage_ready_stages: List[Stage] = []
        setup_stage_candidates: List[Stage] = []
        setup_stage_configs: Dict[str, Config] = {}
        setup_stage_experiment_config: Dict[str, Union[str, int, List[float]]] = {}
        session_stage_monitors: MonitorGroup = {}

        for source_stage, destination_stage in self.history:
            
            previous_history: Dict[str, Any] = self.history[(source_stage, destination_stage)]
            
            if destination_stage == self.source.name:
                setup_stage_configs.update(
                    previous_history.get(
                        'setup_stage_configs',
                        {}
                    )
                )

                execute_config: Config = previous_history['execute_stage_setup_config']
                setup_by = previous_history['execute_stage_setup_by']

                if execute_config.optimized:
                    execute_stage_setup_config = execute_config
                    max_batch_size = execute_config.batch_size
                    execute_stage_setup_by = setup_by

                elif execute_config.batch_size > max_batch_size:
                    execute_stage_setup_config = execute_config
                    max_batch_size = execute_config.batch_size
                    execute_stage_setup_by = setup_by

                execute_hooks: ExecuteHooks = previous_history['execute_stage_setup_hooks']
                for setup_hook in execute_hooks:
                    execute_stage_setup_hooks[setup_hook.name] = setup_hook

                ready_stages = previous_history['setup_stage_ready_stages']

                for ready_stage in ready_stages:
                    if ready_stage not in setup_stage_ready_stages:
                        setup_stage_ready_stages.append(ready_stage)

                stage_candidates: List[Stage] = previous_history['setup_stage_candidates']
                for stage_candidate in stage_candidates:
                    if stage_candidate not in setup_stage_candidates:
                        setup_stage_candidates.append(stage_candidate)

                stage_distributions = previous_history.get('setup_stage_experiment_config')

                if stage_distributions:
                    setup_stage_experiment_config.update(stage_distributions)

                stage_monitors = previous_history.get('session_stage_monitors')
                if stage_monitors:
                    session_stage_monitors.update(stage_monitors)

            streamed_analytics = previous_history.get('execute_stage_streamed_analytics')
            if streamed_analytics:
                execute_stage_streamed_analytics[source_stage].extend(streamed_analytics)

        self.edge_data = {
            'setup_stage_experiment_config': setup_stage_experiment_config,
            'session_stage_monitors': session_stage_monitors,
            'execute_stage_streamed_analytics': execute_stage_streamed_analytics,
            'setup_stage_configs': setup_stage_configs,
            'execute_stage_setup_config': execute_stage_setup_config,
            'execute_stage_setup_hooks': execute_stage_setup_hooks,
            'execute_stage_setup_by': execute_stage_setup_by,
            'setup_stage_ready_stages': setup_stage_ready_stages,
            'setup_stage_candidates': setup_stage_candidates

        }


from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.graphql_http2.action import GraphQLHTTP2Action
from hedra.core.engines.types.graphql_http2.client import MercuryGraphQLHTTP2Client
from hedra.core.engines.types.graphql_http2.result import GraphQLHTTP2Result
from hedra.core.engines.types.common.types import RequestTypes
from hedra.data.serializers.serializer_types.common.base_serializer import BaseSerializer
from typing import List, Dict, Union, Any


class GraphQLHTTP2Serializer(BaseSerializer):

    def __init__(self) -> None:
        super().__init__()

    def action_to_serializable(
        self,
        action: GraphQLHTTP2Action
    ) -> Dict[str, Union[str, List[str]]]:
        serialized_action = super().action_to_serializable(action)

        return {
            **serialized_action,
            'type': RequestTypes.GRAPHQL_HTTP2,
            'url': {
                'full': action.url.full,
                'ip_addr': action.url.ip_addr,
                'socket_config': action.url.socket_config,
                'has_ip_addr': action.url.has_ip_addr
            },
            'method': action.method,
            'headers': action._headers,
            'data': action.data,
            'is_stream': action.is_stream,
            'is_setup': action.is_setup,
            'action_args': action.action_args,
        }
    
    def deserialize_action(
        self,
        action: Dict[str, Any]
    ) -> GraphQLHTTP2Action:
        
        url_config = action.get('url', {})
        metadata = action.get('metadata', {})

        graphql_http2_action = GraphQLHTTP2Action(
            name=action.get('name'),
            url=url_config.get('full'),
            method=action.get('method'),
            headers=action.get('headers'),
            data=action.get('data'),
            user=metadata.get('user'),
            tags=metadata.get('tags', [])
        )

        graphql_http2_action.url.ip_addr = url_config.get('ip_addr')
        graphql_http2_action.url.socket_config = url_config.get('socket_config')
        graphql_http2_action.url.has_ip_addr = url_config.get('has_ip_addr')

        graphql_http2_action.setup()

        return graphql_http2_action
    
    def deserialize_client_config(self, client_config: Dict[str, Any]) -> MercuryGraphQLHTTP2Client:
        return MercuryGraphQLHTTP2Client(
            concurrency=client_config.get('concurrency'),
            timeouts=Timeouts(
                **client_config.get('timeouts', {})
            ),
            reset_connections=client_config.get('reset_sessions')
        )
    
    def result_to_serializable(
        self,
        result: GraphQLHTTP2Result
    ) -> Dict[str, Any]:
        
        serialized_result = super().result_to_serializable(result)

        encoded_headers = {
            str(k.decode()): str(v.decode()) for k, v in result.headers.items()
        }

        body: Union[str, None] = None
        if result.body:
            body = str(result.body.decode())

        return {
            **serialized_result,
            'url': result.url,
            'method': result.method,
            'path': result.path,
            'params': result.params,
            'query': result.query,
            'type': result.type,
            'headers': encoded_headers,
            'body': body,
            'tags': result.tags,
            'user': result.user,
            'error': str(result.error),
            'status': result.status,
            'reason': result.reason,
        }
    
    def deserialize_result(
        self,
        result: Dict[str, Any]
    ) -> GraphQLHTTP2Result:
        deserialized_result = GraphQLHTTP2Result(
            GraphQLHTTP2Action(
                name=result.get('name'),
                url=result.get('url'),
                method=result.get('method'),
                headers=result.get('headers'),
                data=result.get('data'),
                user=result.get('user'),
                tags=result.get('tags', [])      
            ),
            error=Exception(result.get('error'))
        )

        body = result.get('body')
        if isinstance(body, str):
            body = body.encode()

        deserialized_result.body = body
        deserialized_result.status = result.get('status')
        deserialized_result.reason = result.get('reason')
        deserialized_result.params = result.get('params')
        deserialized_result.query = result.get('query')
        deserialized_result.wait_start = result.get('wait_start')
        deserialized_result.start = result.get('start')
        deserialized_result.connect_end = result.get('connect_end')
        deserialized_result.write_end = result.get('write_end')
        deserialized_result.complete = result.get('complete')
        deserialized_result.checks = result.get('checks')

        deserialized_result.type = RequestTypes.GRAPHQL_HTTP2

        return deserialized_result


from __future__ import annotations
import threading
import os
import uuid
from collections import defaultdict
from typing import Any, Dict, List, Optional
from hedra.core.engines.client.time_parser import TimeParser
from hedra.core.hooks.types.base.event_dispatch import EventDispatcher
from hedra.core.hooks.types.internal.decorator import Internal
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.graphs.stages.base.parallel.batch_executor import BatchExecutor
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.hooks.types.base.simple_context import SimpleContext

from hedra.logging import HedraLogger
from hedra.plugins.types.common.plugin import Plugin
from hedra.plugins.types.plugin_types import PluginType



class Stage:
    stage_type=StageTypes
    dependencies: List[Stage]=[]
    all_dependencies: List[Stage]=[]
    next_context: Any = None
    stage_timeout=None
    plugins: Dict[str, Plugin] = {}

    def __init__(self) -> None:
        self.name = self.__class__.__name__
        self.graph_path: str = None
        self.stage_id = str(uuid.uuid4())

        self.state = StageStates.INITIALIZED
        self.next_stage: str = None
        self.hooks: Dict[HookType, List[Hook]] = defaultdict(list)
        self.accepted_hook_types: List[HookType] = []
        self.workers: int = None
        self.worker_id: int = 1
        self.generation_stage_names = []
        self.generation_id = 1
        self.requires_shutdown = False
        self.allow_parallel = False
        self.executor: BatchExecutor = None
        self.plugins_by_type: Dict[PluginType, Dict[str, Plugin]] = defaultdict(dict)
        self.total_pool_cpus: int = 1
        self.stage_retries: int = 0

        self.core_config = {}
        self.context: Optional[SimpleContext] = None
        self.logger: HedraLogger = HedraLogger()
        self.logger.initialize()

        self.graph_name: str = None
        self.graph_id: str = None

        if self.stage_timeout:
            time_parser = TimeParser(self.stage_timeout)

            self.timeout = time_parser.time
        
        else:
            self.timeout = None

        self.internal_hooks = ['run']
        self.dispatcher: EventDispatcher = EventDispatcher()
        self.skip = False

    @Internal()
    async def run(self):
        pass

    @property
    def thread_id(self) -> int:
        return threading.current_thread().ident

    @property
    def process_id(self) -> int:
        return os.getpid()

    @property
    def metadata_string(self):
        return f'Graph - {self.graph_name}:{self.graph_id} - thread:{self.thread_id} - process:{self.process_id} - Stage: {self.name}:{self.stage_id} - '
    
    @Internal()
    async def setup_events(self):

        for event in self.dispatcher.events_by_name.values():
            event.context.update(self.context)
            
            if event.source.context:
                event.source.context.update(self.context)

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing events')

    @Internal()
    def to_copy_dict(self) -> Dict[str, Any]:

        copy_dict = {}

        for attr_name, value in self.__dict__.items():
            if not attr_name.startswith('__'):
                copy_dict[attr_name] = value

        return copy_dict




import asyncio
from typing import (
    TypeVar, 
    List, 
    Dict, 
    Any, 
    Generic, 
    Coroutine
)


A = TypeVar('A')
R = TypeVar('R')


class BaseEngine(Generic[A, R]):

    __slots__ = (
        'waiter'
    )

    def __init__(self) -> None:
        super().__init__()

        self.waiter: asyncio.Future = None

    def config_to_dict(self):
        raise NotImplementedError('Cannot call config_to_dict() on base Engine class.')

    async def wait_for_active_threshold(self):
        if self.waiter is None:
            self.waiter = asyncio.get_event_loop().create_future()
            await self.waiter

    async def execute_before(self, action: A) -> Coroutine[Any, Any, A]:
        action.action_args = {
            'action': action
        }
        
        for before_batch in action.hooks.before:
            results: List[Dict[str, Any]] = await asyncio.gather(*[
                before.call(**{
                    name: value for name, value in action.action_args.items() if name in before.params
                }) for before in before_batch
            ])

            for before_event, result in zip(before_batch, results):
                for data in result.values():
                    if isinstance(result, dict):
                        action.action_args.update(data)

                    else:
                        action.action_args.update({
                            before_event.shortname: data
                        })

        return action

    async def execute_after(self, action: A, response: R) -> Coroutine[Any, Any, R]:
        action.action_args['action'] = action
        action.action_args['result'] = response

        if action.hooks.notify:
            action.action_args.update({
                name: action_or_task.action for name, action_or_task in action.hooks.listeners.items()
            })

        for after_batch in action.hooks.after:
            results: List[Dict[str, Any]] = await asyncio.gather(*[
                after.call(**{
                    name: value for name, value in action.action_args.items() if name in after.params
                }) for after in after_batch
            ])

            for after_event, result in zip(after_batch, results):
                for data in result.values():
                    if isinstance(result, dict):
                        action.action_args.update(data)

                    else:
                        action.action_args.update({
                            after_event.shortname: data
                        })
 
        return response
    
    async def execute_checks(self, action: A, response: R) -> Coroutine[Any, Any, R]:
        action.action_args['action'] = action
        action.action_args['result'] = response

        if response.error:
            return response

        if action.hooks.notify:
            action.action_args.update({
                name: action_or_task.action for name, action_or_task in action.hooks.listeners.items()
            })
            
        for check_batch in action.hooks.checks:
            results: List[Dict[str, Any]] = await asyncio.gather(*[
                check.call(**{
                    name: value for name, value in action.action_args.items() if name in check.params
                }) for check in check_batch
            ])

            for check_event, result in zip(check_batch, results):
                for data in result.values():
                    if isinstance(result, dict):
                        action.action_args.update(data)

                    else:
                        action.action_args.update({
                            check_event.shortname: data
                        })

        return action.action_args.get('result')


from hedra.core.engines.client.config import Config
from hedra.core.engines.types.websocket import (
    WebsocketAction,
    WebsocketResult
)
from hedra.core.engines.types.common.types import RequestTypes
from hedra.data.parsers.parser_types.common.base_parser import BaseParser
from hedra.data.parsers.parser_types.common.result_validator import ResultValidator
from hedra.data.parsers.parser_types.common.parsing import (
    normalize_headers,
    parse_data,
    parse_tags
)
from typing import Any, Coroutine, Dict
from .websocket_action_validator import WebsocketActionValidator


class WebsocketResultParser(BaseParser):

    def __init__(
        self,
        config: Config,
        options: Dict[str, Any]={}
    ) -> None:
        super().__init__(
            WebsocketResultParser.__name__,
            config,
            RequestTypes.GRPC,
            options
        )

    async def parse(
        self, 
        result_data: Dict[str, Any]
    ) -> Coroutine[Any, Any, Coroutine[Any, Any, WebsocketAction]]:
        
        normalized_headers = normalize_headers(result_data)
        content_type = normalized_headers.get('content-type')

        parsed_data = parse_data(
            result_data,
            content_type
        )

        tags_data = parse_tags(result_data)

        generator_action = WebsocketActionValidator(
            engine=result_data.get('engine'),
            name=result_data.get('name'),
            url=result_data.get('url'),
            method=result_data.get('method'),
            headers=normalized_headers,
            params=result_data.get('params'),
            data=parsed_data,
            weight=result_data.get('weight'),
            order=result_data.get('order'),
            user=result_data.get('user'),
            tag=tags_data
        )


        action = WebsocketAction(
            generator_action.name,
            generator_action.url,
            method=generator_action.method,
            headers=generator_action.headers,
            data=generator_action.data,
            user=generator_action.user,
            tags=[
                tag.dict() for tag in generator_action.tags
            ]
        )


        result_validator = ResultValidator(
            error=result_data.get('error'),       
            status=result_data.get('status'),
            reason=result_data.get('reason'),
            params=result_data.get('params'),
            wait_start=result_data.get('wait_start'),
            start=result_data.get('start'),
            connect_end=result_data.get('connect_end'),
            write_end=result_data.get('write_end'),
            complete=result_data.get('complete'),
            checks=result_data.get('checks')
        )

        result = WebsocketResult(
            action,
            error=Exception(result_validator.error) if result_validator.error else None
        )

        result.query = result_validator.query
        result.status = result_validator.status
        result.reason = result_validator.reason
        result.params = result_validator.params
        result.wait_start = result_validator.wait_start
        result.start = result_validator.start
        result.connect_end = result_validator.connect_end
        result.write_end = result_validator.write_end
        result.complete = result_validator.complete
        result.checks = result_validator.checks

        return result






from .generated_graphql_execute_stage import ExecuteGraphQLStage
from .generated_graphql_http2_stage import ExecuteGraphQLHttp2Stage
from .generated_http_execute_stage import ExecuteHTTPStage
from .generated_http2_execute_stage import ExecuteHTTP2Stage
from .generated_http3_execute_stage import ExecuteHTTP3Stage
from .generated_playwright_execute_stage import ExecutePlaywrightStage
from .generated_task_execute_stage import ExecuteTaskStage
from .generated_udp_execute_stage import ExecuteUDPStage
from .generated_websocket_execute_stage import ExecuteWebsocketStage


from hedra.core.engines.types.common.metadata import Metadata
from typing import Dict, List, Union


class MetadataSerializer:

    def __init__(self) -> None:
        self.user: Union[str, None] = None
        self.tags: List[Dict[str, str]] = []

    def serialize_metadata(
        self,
        metadata: Metadata
    ) -> Dict[str, Union[str, List[str]]]:
        return {
            'user': metadata.user,
            'tags': metadata.tags
        }


from .mean import Mean

class Variance:

    __slots__ = (
        'mean',
        't1',
        't1_sum'
    )

    def __init__(self):
        self.mean = Mean()
        self.t1 = 0.0
        self.t1_sum = 0

    def update(self, new_value):
        self.mean.update(new_value)
        self.t1 = self.mean.delta * self.mean.delta_n * self.mean.previous_size
        self.t1_sum += self.t1
        
    def get(self):
        divisor = (self.mean.size - 1.0)
        if divisor == 0:
            divisor = 1

        return self.t1_sum/divisor


from hedra.reporting.types.common.types import ReporterTypes
from pydantic import BaseModel


class PostgresConfig(BaseModel):
    host: str='localhost'
    database: str
    username: str
    password: str
    events_table: str='events'
    metrics_table: str='metrics'
    experiments_table: str='experiments'
    streams_table: str='streams'
    system_metrics_table: str='system_metrics'
    reporter_type: ReporterTypes=ReporterTypes.Postgres

    class Config:
        arbitrary_types_allowed = True


from hedra.versioning.flags.types.unstable.flag import unstable
from typing import Dict
from .trace import Trace
from .tracing_types import (
    RequestHook,
    ResponseHook,
    TraceSignal,
    UrlFilter
)


__name__ = 'hedra'
__version__ = "0.7.12"


@unstable
class TraceSession:

    __slots__ = (
        'url_filter',
        'request_hook',
        'response_hook',
        'trace_signals'
    )

    def __init__(
        self,
        url_filter: UrlFilter = None,
        request_hook: RequestHook = None,
        response_hook: ResponseHook = None,
        **kwargs: Dict[str, TraceSignal]
    ) -> None:
        self.url_filter: UrlFilter = url_filter
        self.request_hook: RequestHook = request_hook
        self.response_hook: ResponseHook = response_hook
        self.trace_signals: Dict[str, TraceSignal] = kwargs

    def create_trace(self) -> Trace:

        return Trace(
            url_filter=self.url_filter,
            request_hook=self.request_hook,
            response_hook=self.response_hook,
            **self.trace_signals
        )








import struct
from hedra.core.engines.types.common.encoder import Encoder
from hedra.core.engines.types.common.protocols.shared.reader import Reader
from hedra.core.engines.types.common.protocols.shared.writer import Writer
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.http2.streams.stream_settings import Settings
from hedra.core.engines.types.http2.windows.window_manager import WindowManager


class Stream:
    READ_NUM_BYTES=65536

    __slots__ = (
        'stream_id',
        'reader',
        'writer',
        'timeouts',
        'max_inbound_frame_size',
        'max_outbound_frame_size',
        'current_outbound_window_size',
        'content_length',
        'expected_content_length',
        'reset_connection',
        'inbound',
        'outbound',
        'closed_by',
        '_authority',
        'frame_buffer',
        'encoder',
        '_remote_settings',
        '_remote_settings_dict',
        'settings_frame',
        'headers_frame',
        'window_frame',
        'connection_data',
        '_STRUCT_HBBBL',
        '_STRUCT_LL',
        '_STRUCT_HL',
        '_STRUCT_LB',
        '_STRUCT_L',
        '_STRUCT_H',
        '_STRUCT_B'
    )

    def __init__(self, stream_id: int, timeouts: Timeouts) -> None:
        self.stream_id = stream_id
        self.reader: Reader = None
        self.writer: Writer = None
        self.timeouts = timeouts
        self.max_inbound_frame_size = 0
        self.max_outbound_frame_size = 0
        self.current_outbound_window_size = 0
        self.content_length = 0
        self.expected_content_length = 0
        self.reset_connection = False
        self.inbound: WindowManager  = None
        self.outbound: WindowManager = None
        self.closed_by = None
        self._authority = None
        self.frame_buffer = None
        self.encoder: Encoder = None
        self._remote_settings = Settings(
            client=False
        )
        self._remote_settings_dict = {
            setting_name: setting_value for setting_name, setting_value in self._remote_settings.items()
        }

        self.settings_frame = None
        self.headers_frame = None
        self.window_frame = None

        self.connection_data = bytearray(b'PRI * HTTP/2.0\r\n\r\nSM\r\n\r\n')

        self._STRUCT_HBBBL = struct.Struct(">HBBBL")
        self._STRUCT_LL = struct.Struct(">LL")
        self._STRUCT_HL = struct.Struct(">HL")
        self._STRUCT_LB = struct.Struct(">LB")
        self._STRUCT_L = struct.Struct(">L")
        self._STRUCT_H = struct.Struct(">H")
        self._STRUCT_B = struct.Struct(">B")

    def write(self, data: bytes):
        self.writer._transport.write(data)

    def read(self, msg_length: int=READ_NUM_BYTES):
        return self.reader.read(msg_length)

    def get_raw_buffer(self) -> bytearray:
        return self.reader._buffer

    def write_window_update_frame(self, stream_id: int=None, window_increment: int=None):

        if stream_id is None:
            stream_id = self.stream_id

        body = self._STRUCT_L.pack(window_increment & 0x7FFFFFFF)
        body_len = len(body)

        type = 0x08

        # Build the common frame header.
        # First, get the flags.
        flags = 0

        header = self._STRUCT_HBBBL.pack(
            (body_len >> 8) & 0xFFFF,  # Length spread over top 24 bits
            body_len & 0xFF,
            type,
            flags,
            stream_id & 0x7FFFFFFF  # Stream ID is 32 bits.
        )

        self.writer.write(header + body)


import asyncio
from typing import List


def _release_waiter(waiter: asyncio.Future, *args):
    if not waiter.done():
        waiter.set_result(None)


async def wait(
    fs: List[asyncio.Future], 
    timeout: float=None
):
    loop = asyncio.get_event_loop()
    """Internal helper for wait().
    The fs argument must be a collection of Futures.
    """
    waiter = loop.create_future()
    timeout_handle = None
    if timeout is not None:
        timeout_handle = loop.call_later(timeout, _release_waiter, waiter)
    counter = len(fs)

    def _on_completion(f: asyncio.Future):
        nonlocal counter
        counter -= 1
        if (counter <= 0 and (not f.cancelled() and
                                                f.exception() is not None)):
            if timeout_handle is not None:
                timeout_handle.cancel()
            if not waiter.done():
                waiter.set_result(None)

    for f in fs:
        f.add_done_callback(_on_completion)

    try:
        await waiter
    finally:
        if timeout_handle is not None:
            timeout_handle.cancel()
        for f in fs:
            f.remove_done_callback(_on_completion)

    done, pending = set(), set()
    for f in fs:
        if f.done():
            done.add(f)
        else:
            pending.add(f)

    return done, pending




from hedra.distributed.models.http import Request
from pydantic import BaseModel
from typing import (
    Callable, 
    Union, 
    Dict, 
    Optional,
    List,
    Literal,
    TypeVar
)
from .base_wrapper import BaseWrapper
from .types import MiddlewareType
from .types import (
    Handler,
    BidirectionalMiddlewareHandler
)


T = TypeVar('T')


class BidirectionalWrapper(BaseWrapper):

    def __init__(
        self,
        name: str,
        handler: Handler,
        middleware_type: MiddlewareType=MiddlewareType.BIDIRECTIONAL,
        methods: Optional[
            List[
                Literal[
                    "GET",
                    "HEAD",
                    "OPTIONS",
                    "POST",
                    "PUT",
                    "PATCH",
                    "DELETE",
                    "TRACE"
                ]
            ]
        ]=None,
        responses: Optional[
            Dict[
                int,
                BaseModel
            ]
        ]=None,
        serializers: Optional[
            Dict[
                int,
                Callable[
                    ...,
                    str
                ]
            ]
        ]=None,
        response_headers: Optional[
            Dict[str, str]
        ]=None
    ) -> None:
        
        super().__init__()

        self.name = name
        self.path  = handler.path
        self.methods: List[
            Literal[
                "GET",
                "HEAD",
                "OPTIONS",
                "POST",
                "PUT",
                "PATCH",
                "DELETE",
                "TRACE"
            ]
        ] = handler.methods

        if methods:
            self.methods.extend(methods)

        self.response_headers: Union[
            Dict[str, str],
            None
        ] = handler.response_headers

        if self.response_headers and response_headers:
            self.response_headers.update(response_headers)

        elif response_headers:
            self.response_headers = response_headers

        self.responses = responses
        self.serializers = serializers
        self.limit = handler.limit

        self.handler = handler
        self.wraps = isinstance(handler, BaseWrapper)

        if self.handler.response_headers and self.response_headers:
            self.handler.response_headers = {}

        self.pre: Optional[BidirectionalMiddlewareHandler] = None
        self.post: Optional[BidirectionalMiddlewareHandler] = None

        self.middleware_type = middleware_type

    async def __call__(
        self, 
        request: Request
    ):

        (request, response, middleware_status), run_next = await self.pre(
            request,
            None,
            None
        )

        if run_next is False:
            return response, middleware_status
        
        if self.wraps:
            result, status = await self.handler(request)
            result.headers.update(response.headers)

        else:
            result, status = await self.handler(request)
            

        (request, response, middleware_status), run_next = await self.post(
            request,
            result,
            status
        )

        if run_next is False:
            return response, middleware_status
        
        return response, status


import uuid
from typing import (
    Any, 
    Dict, 
    List, 
    TypeVar, 
    Generic
)
from .metadata import Metadata
from .hooks import Hooks
from .types import ProtocolMap


A = TypeVar('A')


class BaseAction(Generic[A]):

    __slots__ = ( 
        'action_id'
        'protocols', 
        'name', 
        'is_setup', 
        'metadata', 
        'hooks',
        'event',
        'action_args',
        'mutations'
    )

    def __init__(
        self, 
        name: str=None,
        user: str=None, 
        tags: List[Dict[str, str]] = []
    ) -> None:
        self.action_id = str(uuid.uuid4())
        self.protocols = ProtocolMap()
        self.name = name
        self.is_setup = False
        self.metadata = Metadata(user, tags)
        self.hooks: Hooks[BaseAction] = Hooks()
        self.event = None
        self.action_args: Dict[str, Any] = {}


import asyncio
import functools
import re
import psutil
import uuid
from typing import List, Dict
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)

try:
    import newrelic.agent
    from .newrelic_config import NewRelicConfig
    has_connector=True

except Exception:
    newrelic = None
    NewRelicConfig = None
    has_connector=False


class NewRelic:

    def __init__(self, config: NewRelicConfig) -> None:
        self.config_path = config.config_path
        self.environment = config.environment
        self.registration_timeout = config.registration_timeout
        self.shutdown_timeout = config.shutdown_timeout or 60
        self.newrelic_application_name = config.newrelic_application_name

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self.client = None
        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to NewRelic - Using config at path - {self.config_path} - Environment - {self.environment}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                newrelic.agent.initialize,
                config_file=self.config_path,
                environment=self.environment
            )
        )

        self.client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                newrelic.agent.register_application,
                name=self.newrelic_application_name,
                timeout=self.registration_timeout
            )
        )

        await asyncio.sleep(1)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to NewRelic - Using config at path - {self.config_path} - Environment - {self.environment}')
    
    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to NewRelic')

        metrics_sets: List[SessionMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics)

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.name}:{metrics_set.group}')

            for field, value in metrics_set.record.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.name}:{metrics_set.group}:{field}')

                record_name = f'{metrics_set.name}_{metrics_set.group}_{field}'

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.record_custom_metric,
                        record_name,
                        value
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to NewRelic')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to NewRelic')

        metrics_sets: List[SystemMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')
            
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics)

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.name}:{metrics_set.group}')

            for field, value in metrics_set.record.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.name}:{metrics_set.group}:{field}')

                record_name = f'{metrics_set.name}_{metrics_set.group}_{field}'

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.record_custom_metric,
                        record_name,
                        value
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to NewRelic')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to NewRelic')

        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams - {stage_name}:{stream.stream_set_id}')

            for group_name, group in stream.grouped.items():
                for field, value in group.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams - {stage_name}:{group_name}:{field}')
                    await self._loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            self.client.record_custom_metric,
                            f'{stage_name}_{group_name}_{field}',
                            value
                        )
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to NewRelic')

    async def submit_experiments(self, experiments_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to NewRelic')

        for experiment in experiments_metrics.experiment_summaries:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.record_custom_event,
                    experiment.experiment_name,
                    experiment.record
                )
            )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to NewRelic')

    async def submit_variants(self, experiments_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to NewRelic')

        for variant in experiments_metrics.variant_summaries:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.record_custom_event,
                    variant.variant_name,
                    variant.record
                )
            )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to NewRelic')

    async def submit_mutations(self, experiments_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to NewRelic')

        for mutation in experiments_metrics.mutation_summaries:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.record_custom_event,
                    mutation.mutation_name,
                    mutation.record
                )
            )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to NewRelic')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to NewRelic')

        for event in events:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.record_custom_event,
                    event.name,
                    event.record
                )
            )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to NewRelic')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to NewRelic')
        
        for metrics_set in metrics_sets:  
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for field, value in metrics_set.common_stats.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:common:{field}')
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.record_custom_metric,
                        f'{metrics_set.name}_{field}',
                        value
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to NewRelic')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to NewRelic')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                for field, value in group.stats.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:{group_name}:{field}')
                    await self._loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            self.client.record_custom_metric,
                            f'{metrics_set.name}_{group_name}_{field}',
                            value
                        )
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to NewRelic')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to NewRelic')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metric - {metrics_set.name}:custom:{custom_metric_name}')
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.record_custom_metric,
                        f'{metrics_set.name}_custom_{custom_metric_name}',
                        custom_metric.metric_value
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to NewRelic')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to NewRelic')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                error_message = re.sub(
                    '[^0-9a-zA-Z]+', 
                    '_',
                    error.get(
                        'message'
                    ).lower()
                )

                error_message_metric_name = f'{metrics_set.name}_errors_{error_message}'

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.record_custom_metric,
                        error_message_metric_name,
                        error.get('count')
                    )
                )  
                
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to NewRelic')          

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connectiion to NewRelic')

        await self._loop.run_in_executor(
            self._executor,
            self.client.shutdown
        )

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connectiion to NewRelic')


import asyncio


async def cancel(pending_item: asyncio.Task) -> None:
    pending_item.cancel()
    if not pending_item.cancelled():
        try:
            await pending_item

        except asyncio.CancelledError:
            pass

        except asyncio.IncompleteReadError:
            pass



from .graphql import MercuryGraphQLClient
from .graphql_http2 import MercuryGraphQLHTTP2Client
from .grpc import MercuryGRPCClient
from .http import MercuryHTTPClient
from .http2 import MercuryHTTP2Client
from .http3 import MercuryHTTP3Client
from .playwright import MercuryPlaywrightClient
from .task import MercuryTaskRunner
from .udp import MercuryUDPClient
from .websocket import MercuryWebsocketClient
from .common.types import RequestTypes



registered_engines = {
    RequestTypes.HTTP: lambda concurrency, timeouts, reset_connections: MercuryHTTPClient(
        concurrency=concurrency,
        timeouts=timeouts,
        reset_connections=reset_connections
    ),
    RequestTypes.HTTP2: lambda concurrency, timeouts, reset_connections: MercuryHTTP2Client(
        concurrency=concurrency,
        timeouts=timeouts,
        reset_connections=reset_connections
    ),
    RequestTypes.HTTP3: lambda concurrency, timeouts, reset_connections: MercuryHTTP3Client(
        concurrency=concurrency,
        timeouts=timeouts,
        reset_connections=reset_connections
    ),
    RequestTypes.GRPC: lambda concurrency, timeouts, reset_connections: MercuryGRPCClient(
        concurrency=concurrency,
        timeouts=timeouts,
        reset_connections=reset_connections
    ),
    RequestTypes.GRAPHQL: lambda concurrency, timeouts, reset_connections: MercuryGraphQLClient(
        concurrency=concurrency,
        timeouts=timeouts,
        reset_connections=reset_connections
    ),
    RequestTypes.GRAPHQL_HTTP2: lambda concurrency, timeouts, reset_connections: MercuryGraphQLHTTP2Client(
        concurrency=concurrency,
        timeouts=timeouts,
        reset_connections=reset_connections
    ),
    RequestTypes.PLAYWRIGHT: lambda concurrency, timeouts, reset_connections: MercuryPlaywrightClient(
        concurrency=concurrency,
        timeouts=timeouts,
        reset_connections=reset_connections
    ),
    RequestTypes.TASK: lambda concurrency, timeouts: MercuryTaskRunner(
        concurrency=concurrency,
        timeouts=timeouts
    ),
    RequestTypes.UDP: lambda concurrency, timeouts, reset_connections: MercuryUDPClient(
        concurrency=concurrency,
        timeouts=timeouts,
        reset_connections=reset_connections
    ),
    RequestTypes.WEBSOCKET: lambda concurrency, timeouts, reset_connections: MercuryWebsocketClient(
        concurrency=concurrency,
        timeouts=timeouts,
        reset_connections=reset_connections
    ),
}


from typing import Union
from scipy.stats import cauchy
from .base import BaseDistribution


class CauchyDistribution(BaseDistribution):

    def __init__(
        self, 
        size: int, 
        center: Union[int, float]=0.5,
        randomness: Union[int, float]=0.25
    ) -> None:

        super().__init__(
            size=size,
            center=center,
            randomness=randomness,
            frozen_distribution=cauchy(
                loc=center,
                scale=randomness
            )
        )


from .bidirectional_gzip_compressor import BidirectionalGZipCompressor
from .bidirectional_zstandard_compressor import BidirectionalZStandardCompressor
from .gzip_compressor import GZipCompressor
from .zstandard_compressor import ZStandardCompressor


import asyncio
import socket
from typing import Optional, Callable
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.protocols.shared.reader import Reader
from hedra.core.engines.types.common.protocols.shared.writer import Writer
from .protocol import UDPProtocol
from .quic_protocol import QuicProtocol

try:
    from aioquic.h3.connection import H3_ALPN
    from aioquic.quic.connection import QuicConnection
    from aioquic.quic.configuration import QuicConfiguration

except ImportError:
    H3_ALPN = []
    QuicConnection = object
    QuicConfiguration = object

from hedra.core.engines.types.common.protocols.shared.constants import _DEFAULT_LIMIT


QuicStreamHandler = Callable[[asyncio.StreamReader, asyncio.StreamWriter], None]


class UDPConnection:

    def __init__(self, factory_type: RequestTypes = RequestTypes.HTTP) -> None:
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self.transport: asyncio.DatagramTransport = None
        self.factory_type = factory_type
        self._connection = None
        self.socket: socket.socket = None
        self._writer = None

    async def create_udp(self, socket_config=None, *, limit=_DEFAULT_LIMIT, tls=None):
        
        self.loop = asyncio.get_event_loop()

        family, type_, _, _, address = socket_config

        self.socket = socket.socket(family=family, type=type_)
        
        await self.loop.run_in_executor(None, self.socket.connect, address)

        self.socket.setblocking(False)

        reader = Reader(limit=limit, loop=self.loop)
        reader_protocol = UDPProtocol(reader, loop=self.loop)

        self.transport, _ = await self.loop.create_datagram_endpoint(
            lambda: reader_protocol, 
            sock=self.socket
        )

        # TODO: Enable DTLS - This appears to be a *significant* amount of work
        # but would allow for VOIP testing, etc.

        self._writer = Writer(self.transport, reader_protocol, reader, self.loop)
        
        return reader, self._writer
    
    async def create_http3(
        self, 
        socket_config=None,
        server_name: str=None, 
        configuration: Optional[QuicConfiguration] = None,
        stream_handler: Optional[QuicStreamHandler] = None,
        local_port: int = 0,
    ):
        
        _, _, _, _, address = socket_config
        if len(address) == 2:
            address = ("::ffff:" + address[0], address[1], 0, 0)

        local_host = "::"

        # keep compatibility for Python 3.7 on Windows
        if not hasattr(socket, "IPPROTO_IPV6"):
            socket.IPPROTO_IPV6 = 41

        configuration = QuicConfiguration(
            is_client=True, alpn_protocols=H3_ALPN
        )

        # prepare QUIC connection
        if configuration.server_name is None:
            configuration.server_name = server_name
            
        connection = QuicConnection(
            configuration=configuration, 
            session_ticket_handler=lambda handler: None
        )

        # explicitly enable IPv4/IPv6 dual stack
        self.socket = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
        completed = False
        try:
            self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            self.socket.setblocking(False)
            self.socket.bind((local_host, local_port, 0, 0))
            completed = True
        finally:
            if not completed:
                self.socket.close()
        # connect
        self.loop = asyncio.get_event_loop()
        _, protocol = await self.loop.create_datagram_endpoint(
            lambda: QuicProtocol(
                connection, 
                stream_handler=stream_handler,
                loop=self.loop
            ),
            sock=self.socket,
        )


        protocol.init_connection()

        protocol.connect(address)
        await protocol.wait_connected()

        return protocol
    
    async def close(self):

        try:
            self.transport.close()

        except Exception:
            pass




from gzip import decompress
from hedra.distributed.middleware.base import (
    Middleware,
    MiddlewareType
)
from hedra.distributed.models.http import (
    Request,
    Response
)
from pydantic import BaseModel
from typing import (
    Dict, 
    Union,
    Tuple,
    Callable
)


class BidirectionalGZipDecompressor(Middleware):

    def __init__(
        self, 
        compression_level: int=9,
        serializers: Dict[
            str,
            Callable[
                [
                    Union[
                        Response,
                        BaseModel,
                        str,
                        None
                    ]
                ],
                Union[
                    str,
                    None
                ]
            ]
        ]={}
    ) -> None:
        super().__init__(
            self.__class__.__name__,
            middleware_type=MiddlewareType.BIDIRECTIONAL
        )

        self.compression_level = compression_level
        self.serializers = serializers

    async def __pre__(
        self,
        request: Request,
        response: Union[
            Response,
            BaseModel,
            str,
            None
        ],
        status: int
    ):
        try:

            headers = request.headers
            content_encoding = headers.get(
                'content-encoding',
                headers.get('x-compression-encoding')
            )

            if request.raw != b'' and content_encoding == 'gzip':
                request.content = decompress(
                    request.content
                )

                request_headers = {
                    key: value for key, value in headers.items() if key != 'content-encoding' and key != 'x-compression-encoding'
                }

            return (
                request,
                Response(
                    request.path,
                    request.method,
                    headers=request_headers
                ),
                200
            ), True

        except Exception as e:
            return (
                None,
                Response(
                    request.path,
                    request.method,
                    data=str(e)
                ),
                500
            ), False
        
    async def __post__(
        self, 
        request: Request,
        response: Union[
            Response,
            BaseModel,
            str,
            None
        ],
        status: int
    ) -> Tuple[
        Tuple[Response, int], 
        bool
    ]:
        try:
            
            if response is None:
                return (
                    request,
                    Response(
                        request.path,
                        request.method,
                        data=response
                    ),
                    status
                ), True
            
            elif isinstance(response, str):

                decompressed_data = decompress(
                    response.encode()
                )

                return (
                    request,
                    Response(
                        request.path,
                        request.method,
                        headers={
                            'content-type': 'text/plain'
                        },
                        data=decompressed_data.decode()
                    ),
                    status
                ), True
            
            else:

                headers = response.headers
                content_encoding = headers.get(
                    'content-encoding',
                    headers.get('x-compression-encoding')
                )
                
                if content_encoding == 'gzip':

                    serialized = self.serializers[request.path](response)
                    decompressed_data = decompress(
                        serialized
                    )

                    headers.pop(
                        'content-encoding',
                        headers.pop(
                            'x-compression-encoding',
                            None
                        )
                    )

                    return (
                        request,
                        Response(
                            request.path,
                            request.method,
                            headers=headers,
                            data=decompressed_data.decode()
                        ),
                        status
                    ), True
       
                return (
                    response,
                    status
                ), True
            
        except KeyError:
            return (
                request,
                Response(
                    request.path,
                    request.method,
                    data=f'No serializer for {request.path} found.'
                ),
                500
            ), False

        except Exception as e:
            return (
                request,
                Response(
                    request.path,
                    request.method,
                    data=str(e)
                ),
                500
            ), False


import os
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat,
    IPvAnyAddress
)
from typing import (
    Dict, 
    Union,
    Callable,
    Literal
)


PrimaryType = Union[str, int, float, bytes, bool]


class Env(BaseModel):
    MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_REJECTION_SENSITIVITY: StrictFloat=2
    MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_WINDOW: StrictStr='1m'
    MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_THRESHOLD: Union[StrictInt, StrictFloat]=0.2
    MERCURY_SYNC_HTTP_HANDLER_TIMEOUT: StrictStr='1m'
    MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY: Literal[
        "none", 
        "global", 
        "endpoint", 
        "ip",
        "ip-endpoint",
        "custom"
    ]="none"
    MERCURY_SYNC_HTTP_RATE_LIMITER_TYPE: Literal[
        "adaptive",
        "cpu-adaptive",
        "leaky-bucket",
        "rate-adaptive",
        "sliding-window",
        "token-bucket",
    ]="sliding-window"
    MERCURY_SYNC_HTTP_CORS_ENABLED: StrictBool=False
    MERCURY_SYNC_HTTP_MEMORY_LIMIT: StrictStr='512mb'
    MERCURY_SYNC_HTTP_CPU_LIMIT: Union[StrictFloat, StrictInt]=50
    MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF_RATE: StrictInt=10
    MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF: StrictStr='1s'
    MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD: StrictStr='1s'
    MERCURY_SYNC_HTTP_RATE_LIMIT_REQUESTS: StrictInt=100
    MERCURY_SYNC_HTTP_RATE_LIMIT_DEFAULT_REJECT: StrictBool=True
    MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION: StrictBool=False
    MERCURY_SYNC_USE_HTTP_SERVER: StrictBool=False
    MERCURY_SYNC_USE_HTTP_AND_TCP_SERVERS: StrictBool=False
    MERCURY_SYNC_USE_UDP_MULTICAST: StrictBool=False
    MERCURY_SYNC_TCP_CONNECT_RETRIES: StrictInt=3
    MERCURY_SYNC_CLEANUP_INTERVAL: StrictStr='0.5s'
    MERCURY_SYNC_MAX_CONCURRENCY: StrictInt=2048
    MERCURY_SYNC_AUTH_SECRET: StrictStr
    MERCURY_SYNC_MULTICAST_GROUP: IPvAnyAddress='224.1.1.1'
    MERCURY_SYNC_LOGS_DIRECTORY: StrictStr=os.getcwd()
    MERCURY_SYNC_REQUEST_TIMEOUT: StrictStr='30s'
    MERCURY_SYNC_LOG_LEVEL: StrictStr='info'

    @classmethod
    def types_map(self) -> Dict[str, Callable[[str], PrimaryType]]:
        return {
            'MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_REJECTION_SENSITIVITY': float,
            'MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_WINDOW': str,
            'MERCURY_SYNC_HTTP_HANDLER_TIMEOUT': str,
            'MERCURY_SYNC_USE_UDP_MULTICAST': lambda value: True if value.lower() == 'true' else False,
            'MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_THRESHOLD': float,
            'MERCURY_SYNC_HTTP_CORS_ENABLED': lambda value: True if value.lower() == 'true' else False,
            'MERCURY_SYNC_HTTP_MEMORY_LIMIT': str,
            'MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF_RATE': int,
            'MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF': str,
            'MERCURY_SYNC_HTTP_CPU_LIMIT': float,
            'MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY': str,
            'MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD': str,
            'MERCURY_SYNC_USE_TCP_SERVER': lambda value: True if value.lower() == 'true' else False,
            'MERCURY_SYNC_HTTP_RATE_LIMIT_REQUESTS': int,
            'MERCURY_SYNC_HTTP_RATE_LIMIT_DEFAULT_REJECT': lambda value: True if value.lower() == 'true' else False,
            'MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION': lambda value: True if value.lower() == 'true' else False,
            'MERCURY_SYNC_USE_HTTP_SERVER': lambda value: True if value.lower() == 'true' else False,
            'MERCURY_SYNC_TCP_CONNECT_RETRIES': int,
            'MERCURY_SYNC_CLEANUP_INTERVAL': str,
            'MERCURY_SYNC_MAX_CONCURRENCY': int,
            'MERCURY_SYNC_AUTH_SECRET': str,
            'MERCURY_SYNC_MULTICAST_GROUP': str,
            'MERCURY_SYNC_LOGS_DIRECTORY': str,
            'MERCURY_SYNC_REQUEST_TIMEOUT': str,
            'MERCURY_SYNC_LOG_LEVEL': str
        }


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    SnowflakeConfig
)


class SubmitSnowflakeResultsStage(Submit):
    config=SnowflakeConfig(
        username=os.getenv('SNOWFLAKE_USERNAME', ''),
        password=os.getenv('SNOWFLAKE_PASSWORD', ''),
        organization_id=os.getenv('SNOWFLAKE_ORGANIZATION', ''),
        account_id=os.getenv('SNOWFLAKE_ACCOUNT_ID', ''),
        private_key=os.getenv('SNOWFLAKE_PRIVATE_KEY', ''),
        warehouse=os.getenv('SNOWFLAKE_WAREHOUSE', ''),
        database='results',
        database_schema='results',
        events_table='events',
        metrics_table='metrics'
    )


import time 
import uuid
import asyncio
from typing import Any, Coroutine, Optional, Union
from hedra.core.engines.types.http2 import MercuryHTTP2Client
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.types.tracing.trace_session import (
    TraceSession, 
    Trace
)

from .action import GRPCAction
from .result import GRPCResult


class MercuryGRPCClient(MercuryHTTP2Client[GRPCAction, GRPCResult]):

    def __init__(
        self, 
        concurrency: int = 10 ** 3, 
        timeouts: Timeouts = None, 
        reset_connections: bool=False,
        tracing_session: Optional[TraceSession]=None
    ) -> None:
        super(
            MercuryGRPCClient,
            self
        ).__init__(
            concurrency=concurrency, 
            timeouts=timeouts, 
            reset_connections=reset_connections,
            tracing_session=tracing_session
        )

        self.session_id = str(uuid.uuid4())

    async def execute_prepared_request(self, action: GRPCAction) -> Coroutine[Any, Any, GRPCResult]:

        trace: Union[Trace, None] = None
        if self.tracing_session:
            trace = self.tracing_session.create_trace()
            await trace.on_request_start(action)

        response = GRPCResult(action)
        response.wait_start = time.monotonic()
        self.active += 1

        if trace and trace.on_connection_queued_start:
            await trace.on_connection_queued_start(
                trace.span,
                action,
                response
            )
        
        async with self.sem:

            if trace and trace.on_connection_queued_end:
                await trace.on_connection_queued_end(
                    trace.span,
                    action,
                    response
                )

            pipe = self.pool.pipes.pop()
            connection = self.pool.connections.pop()
        
            try:

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action: GRPCAction = await self.execute_before(action)
                    action.setup()

                response.start = time.monotonic()

                if trace and trace.on_connection_create_start:
                    await trace.on_connection_create_start(
                        trace.span,
                        action,
                        response
                    )

                stream = await connection.connect(
                    action.url.hostname,
                    action.url.ip_addr,
                    action.url.port,
                    action.url.socket_config,
                    ssl=action.ssl_context,
                    timeout=self.timeouts.connect_timeout
                )

                stream.encoder = action.hpack_encoder
     
                response.connect_end = time.monotonic()

                if trace and trace.on_connection_create_end:
                    await trace.on_connection_create_end(
                        trace.span,
                        action,
                        response
                    )

                pipe.send_request_headers(action, stream)

                if trace and trace.on_request_headers_sent:
                    await trace.on_request_headers_sent(
                        trace.span,
                        action,
                        response
                    )
  
                if action.encoded_data is not None:
                    await pipe.submit_request_body(action, stream)

                response.write_end = time.monotonic()

                if action.encoded_data and trace and trace.on_request_data_sent:
                        await trace.on_request_data_sent(
                            trace.span,
                            action,
                            response
                        )

                await asyncio.wait_for(
                    pipe.receive_response(
                        action,
                        response, 
                        stream,
                        trace
                    ), 
                    timeout=self.timeouts.total_timeout
                )

                response.complete = time.monotonic()

                if action.hooks.after:
                    response: GRPCResult = await self.execute_after(action, response)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel.call(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()      

                self.pool.pipes.append(pipe)
                self.pool.connections.append(connection)
                
            except Exception as e:
                response.complete = time.monotonic()
                response._status = 400
                response.error = str(e)

                self.pool.reset()

                if trace and trace.on_request_exception:
                    await trace.on_request_exception(response)

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            if trace and trace.on_request_end:
                await trace.on_request_end(response)

            return response



import statistics
import numpy
import uuid
from hedra.reporting.metric.metric_types import MetricType
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from typing import List, Dict, Union, Callable


CalculationMethod = Callable[[Dict[str,List[Union[int, float]]]], Dict[str, Union[int, float]]]


class StageStreamsSet:

    def __init__(self, 
        stage: str,
        stage_streams: List[StreamAnalytics]
    ) -> None:
        self.stage = stage
        self.stream_set_id = uuid.uuid4()
        self.interval_completion_rates: List[float] = []
        self.interval_completed_counts: List[int] = []
        self.interval_succeeded_counts: List[int] = []
        self.interval_failed_counts: List[int] = []
        self.interval_batch_timings: List[float] = []

        for stream in stage_streams:
            self.interval_completion_rates.extend(stream.interval_completion_rates)
            self.interval_completed_counts.extend(stream.interval_completed_counts)
            self.interval_succeeded_counts.extend(stream.interval_succeeded_counts)
            self.interval_failed_counts.extend(stream.interval_failed_counts)
            self.interval_batch_timings.extend(stream.interval_batch_timings)

        self._stat_types: Dict[str, CalculationMethod] = {
            'median': self._calculate_median,
            'mean': self._calculate_mean,
            'max': self._calculate_max,
            'min': self._calculate_min,
            'stdev': self._calculate_stdev,
            'variance': self._calculate_var,
        }

        self.quantiles = [
            10,
            20,
            25,
            30,
            40,
            50,
            60,
            70,
            75,
            80,
            90,
            95,
            99
        ]

        self.stats: Dict[str, List[Union[int, float]]] = {
            'completed': self.interval_completed_counts,
            'succeeded': self.interval_succeeded_counts,
            'failed': self.interval_failed_counts,
            'batch_time': self.interval_batch_timings,
            'completion_rate': self.interval_completion_rates
        }


    @property
    def types_map(self) -> Dict[str, MetricType]:
        return {
            'completed': MetricType.COUNT,
            'succeeded': MetricType.COUNT,
            'failed': MetricType.COUNT,
            'batch_time': MetricType.SAMPLE,
            'completion_rate': MetricType.RATE
        }

    @property
    def grouped(self) -> Dict[str, Dict[str, Union[int, float]]]:

        
        grouped: Dict[str, Dict[str, Union[int, float]]] = {}

        for group_name, group_stats in self.stats.items():

            metrics: Dict[str, Union[int, float]] = {
                stat_name: calculation_method(
                    group_stats
                ) for stat_name, calculation_method in self._stat_types.items()
            }

            metrics.update(
                self._calculate_quantiles(group_stats)
            )

            grouped[group_name] = metrics

        return grouped

    @property
    def record(self) -> Dict[str, Union[int, float]]:

        record_stats: Dict[str, Dict[str, Union[int, float]]] = {}

        for group_name, group_stats in self.stats.items():
            for stat_name, calculation_method in self._stat_types.items():
                record_key = f'{stat_name}_{group_name}'

                record_stats[record_key] = calculation_method(group_stats)


        median_batch_time = record_stats['median_batch_time']
        if median_batch_time == 0:
            median_batch_time = 1

        actions_per_second = record_stats['median_completed']/median_batch_time
        actions_per_second_succeeded = record_stats['median_succeeded']/median_batch_time
        actions_per_second_failed = record_stats['median_failed']/median_batch_time

        record_stats.update({
            'actions_per_second': actions_per_second,
            'actions_per_second_succeeded': actions_per_second_succeeded,
            'actions_per_second_failed': actions_per_second_failed
        })

        return record_stats
    
    def _calculate_mean(self, stats:  List[Union[int, float]]) -> float:
        return statistics.mean(stats)

    def _calculate_median(self, stats: List[Union[int, float]]) -> Union[int, float]:
        return statistics.median(stats)

    def _calculate_min(self, stats:List[Union[int, float]]) -> Union[int, float]:
        return min(stats)
    
    def _calculate_max(self, stats: List[Union[int, float]]) -> Union[int ,float]:
        return max(stats)
    
    def _calculate_stdev(self, stats: List[Union[int, float]]):
        return statistics.stdev(stats)
    
    def _calculate_var(self, stats: List[Union[int, float]]) -> Union[int, float]:
        return statistics.variance(stats)
    
    def _calculate_quantiles(self, stats: List[Union[int, float]]) -> Dict[str, Union[int, float]]:
        
        quantile_stats: Dict[str, Union[int, float]] = {}

        for quantile in self.quantiles:
            stat_key = f'quantile_{quantile}th'

            quantile_stats[stat_key] = numpy.quantile(
                stats,
                round(
                    quantile/100,
                    2
                )
            )

        return quantile_stats



import asyncio
import csv
import psutil
import uuid
import os
import functools
import pathlib
import signal
from typing import (
    List, 
    TextIO, 
    Dict, 
    Any,
    Coroutine
)
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from .csv_connector_config import CSVConnectorConfig
from .csv_load_validator import CSVLoadValidator
has_connector = True


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop, 
    csv_file: TextIO
): 
    try:
        csv_file.close()
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass
    

class CSVConnector:
    connector_type=ConnectorType.CSV

    def __init__(
        self, 
        config: CSVConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        self.filepath = config.filepath
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.stage = stage
        self.parser_config = parser_config

        self.logger = HedraLogger()
        self.logger.initialize()

        self._csv_reader: csv.DictReader = None
        self.csv_file: TextIO = None

        self._loop: asyncio.AbstractEventLoop = None
        self.file_mode = config.file_mode
        self.headers = config.headers

        self.parser = Parser()

    async def connect(self):
        self._loop = asyncio._get_running_loop()
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping connect')

        if self.filepath[:2] == '~/':
            user_directory = pathlib.Path.home()
            self.filepath = os.path.join(
                user_directory,
                self.filepath[2:]
            )

        self.filepath = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.abspath,
                self.filepath
            )
        )
        
        if self.csv_file is None:
            self.csv_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.filepath,
                    self.file_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.csv_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening from file - {self.filepath}')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)
    
    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
    
        actions = await self.load_data()

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data()

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        
        csv_options = CSVLoadValidator(**options)
        headers = csv_options.headers

        if headers is None and self.headers:
            headers = self.headers
        
        if self._csv_reader is None:
            self._csv_reader = csv.DictReader(
                self.filepath, 
                fieldnames=headers
            )

        return await self._loop.run_in_executor(
            self._executor,
            self._load_data
        )
    
    def _load_data(self):
        return [ row for row in self._csv_reader ]
    
    async def close(self):

        await self._loop.run_in_executor(
            self._executor,
            self.csv_file.close
        )

        self._executor.shutdown(cancel_futures=True)     
        



import os
import psutil
import itertools
from hedra.monitoring.base.monitor import BaseMonitor


class MemoryMonitor(BaseMonitor):

    def __init__(self) -> None:
        super().__init__()
        self.total_memory = psutil.virtual_memory().total

    def update_monitor(self, monitor_name: str):
        process = psutil.Process(os.getpid())
        mem_info = process.memory_info()

        self.active[monitor_name].append(mem_info.rss)

    def aggregate_worker_stats(self):
        monitor_stats = self._collect_worker_stats()

        for monitor_name, metrics in monitor_stats.items():

            self.collected[monitor_name] = [
                sum(cpu_usage) for cpu_usage in itertools.zip_longest(
                    *metrics,
                    fillvalue=0
                )
            ]
            
            self.stage_metrics[monitor_name] = self.collected[monitor_name]



class NotSetError(Exception):

    def __init__(self, path: str, missing_item: str, flag: str) -> None:
        super().__init__(
            f'\n\nError - {missing_item} found for graph repostiory at {path}.\nSet the initial value for {missing_item} by providing the {flag} option once or check your path.\n'
        )


import uuid
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from .mongodb_config import MongoDBConfig


try:
    from motor.motor_asyncio import AsyncIOMotorClient
    has_connector = True

except Exception:
    AsyncIOMotorClient = None
    has_connector = False



class MongoDB:

    def __init__(self, config: MongoDBConfig) -> None:
        self.host = config.host
        self.username = config.username
        self.password = config.password
        self.database_name = config.database

        self.events_collection = config.events_collection
        self.metrics_collection = config.metrics_collection
        self.streams_collection = config.streams_collection

        self.experiments_collection = config.experiments_collection
        self.variants_collection = f'{config.experiments_collection}_variants'
        self.mutations_collection = f'{config.experiments_collection}_mutations'

        self.session_system_metrics_collection = f'{config.system_metrics_collection}_session'
        self.stage_system_metrics_collection = f'{config.system_metrics_collection}_stage'

        self.shared_metrics_collection = f'{self.metrics_collection}_common'
        self.errors_collection = f'{self.metrics_collection}_errors'
        self.custom_metrics_collection = f'{self.metrics_collection}_custom'

        self.connection: AsyncIOMotorClient = None
        self.database = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to MongoDB instance at - {self.host} - Database: {self.database_name}')

        if self.username and self.password:
            connection_string = f'mongodb://{self.username}:{self.password}@{self.host}/{self.database_name}'
        
        else:
            connection_string = f'mongodb://{self.host}/{self.database_name}'

        self.connection = AsyncIOMotorClient(connection_string)
        self.database = self.connection[self.database_name]

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to MongoDB instance at - {self.host} - Database: {self.database_name}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to Bucket - {self.session_system_metrics_collection}')

        metrics_sets: List[SessionMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics.record)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics.record)

        await self.database[self.metrics_collection].insert_many(metrics_sets)

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to Bucket - {self.session_system_metrics_collection}')

        metrics_sets: List[SystemMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')
            
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics.record)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics.record)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics.record)

        await self.database[self.metrics_collection].insert_many(metrics_sets)

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Bucket - {self.streams_collection}')

        records = []
        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams - {stage_name}:{stream.stream_set_id}')
            
            for group_name, group in stream.grouped.items():
                records.append({
                    'name': f'{stage_name}_streams',
                    'stage': stage_name,
                    'group': group_name,
                    **group
                })

        await self.database[self.metrics_collection].insert_many(records)

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet): 
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Collection - {self.experiments_collection}')
        await self.database[self.events_collection].insert_many(experiment_metrics.experiments)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to Collection - {self.experiments_collection}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet): 
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Collection - {self.variants_collection}')
        await self.database[self.events_collection].insert_many(experiment_metrics.variants)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to Collection - {self.variants_collection}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet): 
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Collection - {self.mutations_collection}')
        await self.database[self.events_collection].insert_many(experiment_metrics.mutations)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to Collection - {self.mutations_collection}')

    async def submit_events(self, events: List[BaseProcessedResult]): 
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Collection - {self.events_collection}')
        await self.database[self.events_collection].insert_many(
            [event.record for event in events]
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Collection - {self.events_collection}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Bucket - {self.shared_metrics_collection}')
        await self.database[self.shared_metrics_collection].insert_many([
            {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'common',
                **metrics_set.common_stats
            } for metrics_set in metrics_sets
        ])

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Bucket - {self.shared_metrics_collection}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Bucket - {self.metrics_collection}')

        records = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for group_name, group in metrics_set.groups.items():
                records.append({
                    'group': group_name,
                    **group.record
                })

        await self.database[self.metrics_collection].insert_many(records)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Bucket - {self.metrics_collection}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        records = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            records.append({
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'custom',
                **{
                    custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                }
            })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Bucket - {self.custom_metrics_collection}')

        await self.database[self.custom_metrics_collection].insert_many(records)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Bucket - {self.custom_metrics_collection}')

    async def submit_errors(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Bucket - {self.errors_collection}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            await self.database[self.errors_collection].insert_many([
                {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'error_message': error.get('message'),
                    'error_count': error.get('count')
                } for error in metrics_set.errors
            ])

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Bucket - {self.errors_collection}')

    async def close(self):
        await self.connection.close()
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')


import asyncio
import time
import uuid
from typing import (
    Dict, 
    Union, 
    Coroutine, 
    Any, 
    Optional
)
from hedra.core.engines.types.common.base_engine import BaseEngine
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.concurrency import Semaphore
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.engines.types.tracing.trace_session import (
    TraceSession, 
    Trace
)

from .task import Task
from .result import TaskResult


class MercuryTaskRunner(BaseEngine[Task, Union[BaseResult, TaskResult]]):

    __slots__ = (
        'pool',
        'session_id',
        'timeouts',
        'concurrency',
        'registered',
        'sem',
        'active',
        'waiter',
        'closed',
        'tracing_session'
    )

    def __init__(
        self, 
        concurrency: int=10**3, 
        timeouts: Timeouts = Timeouts(),
        tracing_session: Optional[TraceSession]=None
    ) -> None:
        super(
            MercuryTaskRunner,
            self
        ).__init__()
        
        self.session_id = str(uuid.uuid4())

        self.timeouts = timeouts
        self.concurrency = concurrency

        self.registered: Dict[str, Task] = {}
        self.closed = False
        self.pool = SimpleContext()
        self.pool.size = concurrency
        self.pool.reset_connections = False
        self.pool.create_pool = lambda: None

        self.sem = asyncio.Semaphore(value=concurrency)
        self.active = 0
        self.waiter = None
        self.tracing_session: TraceSession = tracing_session

    async def set_pool(self, concurrency: int):
        self.pool = SimpleContext()
        self.pool.size = concurrency
        self.pool.reset_connections = False
        self.pool.create_pool = lambda: None

        self.sem = asyncio.Semaphore(value=concurrency)

    def extend_pool(self, increased_capacity: int):
        self.concurrency += increased_capacity
        self.sem = Semaphore(self.concurrency)

    def shrink_pool(self, decrease_capacity: int):
        self.concurrency -= decrease_capacity
        self.sem = Semaphore(self.concurrency)

    async def execute_prepared_request(self, task: Task) -> Coroutine[Any, Any, Union[BaseResult, TaskResult]]:

        trace: Union[Trace, None] = None
        if self.tracing_session:
            trace = self.tracing_session.create_trace()
            await trace.on_request_start(task)

        result = None
        wait_start = time.monotonic()
        self.active += 1

        if trace and trace.on_connection_queued_start:
            await trace.on_connection_queued_start(
                trace.span,
                task,
                result
            )

        start = 0
 
        async with self.sem:

            if trace and trace.on_connection_queued_end:
                await trace.on_connection_queued_end(
                    trace.span,
                    task,
                    result
                )
            
            try:

                if task.hooks.listen:
                    event = asyncio.Event()
                    task.hooks.channel_events.append(event)
                    await event.wait()
                
                if task.hooks.before:
                    task = await self.execute_before(task)

                start = time.monotonic()

                if trace and trace.on_task_start:
                    await trace.on_task_start(
                        trace.span,
                        task,
                        result
                    )

                result: BaseResult = await task.execute(**{
                    name: value for name, value in task.task_args.items() if name in task.params
                })
                
                result.name = task.name
                result.source = task.source
                result.user = task.metadata.user
                result.tags = task.metadata.tags
                result.checks = task.hooks.checks
                result.wait_start = wait_start
                result.start = start
                result.complete = time.monotonic()

                if task.hooks.after:
                    result = await self.execute_after(task, result)

                if task.hooks.notify:
                    for listener in task.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()   

            except Exception as e:
                result = TaskResult(task)

                result.wait_start = wait_start
                result.start = start
                result.complete = time.monotonic()
                result.error = str(e)

                if trace and trace.on_task_error:
                    await trace.on_task_error(
                        trace.span,
                        task,
                        result
                    )

            self.active -= 1
            if self.waiter and self.active <= self.concurrency:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            if trace and trace.on_task_end:
                await trace.on_task_end(
                    trace.span,
                    task,
                    result
                )

            return result

    async def close(self):
        pass



class InvalidTermError(Exception):

    def __init__(
        self,
        entry_id: int,
        entry_term: int,
        expected_term: int
    ) -> None:
        super().__init__(
            f'Log entry - {entry_id} - provided invalid term - {entry_term} - Expected term - {expected_term}'
        )


import asyncio
import uuid
from typing import (
    List, 
    Dict, 
    Any,
    Coroutine
)
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from hedra.logging import HedraLogger
from .mongodb_connector_config import MongoDBConnectorConfig


try:
    from motor.motor_asyncio import AsyncIOMotorClient
    has_connector = True

except Exception:
    AsyncIOMotorClient = None
    has_connector = False


class MongoDBConnector:
    connector_type=ConnectorType.MongoDB

    def __init__(
        self, 
        config: MongoDBConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        self.host = config.host
        self.username = config.username
        self.password = config.password
        self.database_name = config.database
        self.stage = stage
        self.parser_config = parser_config

        self.collection = config.collection
        self.connection: AsyncIOMotorClient = None
        self.database = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()
        self.parser = Parser()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to MongoDB instance at - {self.host} - Database: {self.database_name}')

        if self.username and self.password:
            connection_string = f'mongodb://{self.username}:{self.password}@{self.host}/{self.database_name}'
        
        else:
            connection_string = f'mongodb://{self.host}/{self.database_name}'

        self.connection = AsyncIOMotorClient(connection_string)
        self.database = self.connection[self.database_name]

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to MongoDB instance at - {self.host} - Database: {self.database_name}')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        
        actions: List[Dict[str, Any]] = await self.load_data(
            options=options
        )

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])

    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data(
            options=options
        )

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })

    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        return await self.database[self.collection].find(
            limit=options.get('limit')
        )
    
    async def close(self):
        await self.connection.close()




from .core.hooks import (
    action,
    channel,
    check,
    condition,
    context,
    depends,
    event,
    metric,
    load,
    save,
    task,
    transform,
)

from hedra.core.engines.client import TracingConfig
from hedra.core.engines.types.tracing.trace import Trace
from hedra.core.experiments import (
    Experiment,
    Variant
)

from hedra.core.experiments.mutations import (
    DeformHeader,
    InjectHeader,
    InjectJunkData,
    InjectPing,
    SmuggleRequest
)

from .core.graphs.stages import (
    Act,
    Analyze,
    Execute,
    Optimize,
    Setup,
    Submit,
)
from hedra.core.graphs.stages.optimize.optimization.parameters import Parameter

from .reporting import (
    AWSLambdaConfig,
    AWSTimestreamConfig,
    BigQueryConfig,
    BigTableConfig,
    CassandraConfig,
    CloudwatchConfig,
    CosmosDBConfig,
    CSVConfig,
    DatadogConfig,
    DogStatsDConfig,
    GoogleCloudStorageConfig,
    GraphiteConfig,
    HoneycombConfig,
    InfluxDBConfig,
    JSONConfig,
    KafkaConfig,
    MongoDBConfig,
    MySQLConfig,
    NetdataConfig,
    NewRelicConfig,
    PostgresConfig,
    PrometheusConfig,
    RedisConfig,
    S3Config,
    SnowflakeConfig,
    SQLiteConfig,
    StatsDConfig,
    TelegrafConfig,
    TelegrafStatsDConfig,
    TimescaleDBConfig,
    XMLConfig
)

from .data.connectors.connector import (
    AWSLambdaConnectorConfig,
    BigTableConnectorConfig,
    CassandraConnectorConfig,
    CosmosDBConnectorConfig,
    Connector,
    CSVConnectorConfig,
    GoogleCloudStorageConnectorConfig,
    HARConnectorConfig,
    JSONConnectorConfig,
    KafkaConnectorConfig,
    MongoDBConnectorConfig,
    MySQLConnectorConfig,
    PostgresConnectorConfig,
    RedisConnectorConfig,
    S3ConnectorConfig,
    SnowflakeConnectorConfig,
    SQLiteConnectorConfig,
    XMLConnectorConfig
)



from pydantic import (
    BaseModel,
    StrictInt,
    StrictStr
)
from typing import (
    Dict, 
    Union,
    Callable
)


PrimaryType = Union[str, int, float, bytes, bool]


class ReplicationEnv(BaseModel):
    MERCURY_SYNC_RAFT_ELECTION_MAX_TIMEOUT: StrictStr='30s'
    MERCURY_SYNC_RAFT_ELECTION_POLL_INTERVAL: StrictStr='1s'
    MERCURY_SYNC_RAFT_LOGS_UPDATE_POLL_INTERVAL: StrictStr='1s'
    MERCURY_SYNC_RAFT_REGISTRATION_TIMEOUT: StrictStr='15s'
    MERCURY_SYNC_RAFT_EXPECTED_NODES: StrictInt=3
    MERCURY_SYNC_RAFT_LOGS_PRUNE_MAX_AGE: StrictStr='1h'
    MERCURY_SYNC_RAFT_LOGS_PRUNE_MAX_COUNT: StrictInt=1000
   
    @classmethod
    def types_map(self) -> Dict[str, Callable[[str], PrimaryType]]:
        return {
            'MERCURY_SYNC_RAFT_ELECTION_MAX_TIMEOUT': str,
            'MERCURY_SYNC_RAFT_ELECTION_POLL_INTERVAL': str,
            'MERCURY_SYNC_RAFT_LOGS_UPDATE_POLL_INTERVAL': str,
            'MERCURY_SYNC_RAFT_REGISTRATION_TIMEOUT': str,
            'MERCURY_SYNC_RAFT_EXPECTED_NODES': int,
            'MERCURY_SYNC_RAFT_LOGS_PRUNE_MAX_AGE': str,
            'MERCURY_SYNC_RAFT_LOGS_PRUNE_MAX_COUNT': int
        }


from enum import Enum


class MutationType(Enum):
    DEFORM_HEADER='DEFORM_HEADER'
    INJECT_HEADER='INJECT_HEADER'
    INJECT_JUNK_DATA='INJECT_JUNK_DATA'
    INJECT_PING='INJECT_PING'
    SMUGGLE_REQUEST='SMUGGLE_REQUEST'


from .action.decorator import action
from .channel.decorator import channel
from .check.decorator import check
from .condition.decorator import condition
from .context.decorator import context
from .depends.decorator import depends
from .event.decorator import event
from .internal.decorator import Internal  
from .metric.decorator import metric
from .load.decorator import load
from .save.decorator import save
from .task.decorator import task
from .transform.decorator import transform


from __future__ import annotations
import asyncio
import ipaddress
import psutil
import socket
import ssl
import zstandard
from collections import deque, defaultdict
from hedra.distributed.env import Env
from hedra.distributed.connection.base.connection_type import ConnectionType
from hedra.distributed.models.http import (
    HTTPMessage,
    HTTPRequest,
    Response,
    Request
)
from hedra.distributed.rate_limiting import Limiter
from pydantic import BaseModel
from typing import (
    Tuple, 
    Union, 
    Optional, 
    Deque, 
    Dict, 
    List, 
    Callable
)
from .mercury_sync_tcp_connection import MercurySyncTCPConnection
from .protocols import MercurySyncTCPClientProtocol


class MercurySyncHTTPConnection(MercurySyncTCPConnection):

    def __init__(
        self, 
        host: str, 
        port: int,
        instance_id: int, 
        env: Env,
    ) -> None:
        super().__init__(
            host, 
            port, 
            instance_id, 
            env
        )

        self._waiters: Deque[asyncio.Future] = deque()
        self._connections: Dict[str, List[asyncio.Transport]] = defaultdict(list)
        self._http_socket: Union[socket.socket, None] = None
        self._hostnames: Dict[Tuple[str, int], str] = {}
        self._max_concurrency = env.MERCURY_SYNC_MAX_CONCURRENCY

        self.connection_type = ConnectionType.HTTP
        self._is_server = env.MERCURY_SYNC_USE_HTTP_SERVER
        self._use_encryption = env.MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION

        self._supported_handlers: Dict[str, Dict[str, str]] = defaultdict(dict)
        self._response_parsers: Dict[
            Tuple[str, int],
            Callable[
                [BaseModel],
                str
            ]
        ] = {}

        self._middleware_enabled: Dict[str, bool] = {}

        self._limiter = Limiter(env)

        self._backoff_sem: Union[asyncio.Semaphore, None] = None

        rate_limit_strategy = env.MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY
        self._rate_limiting_enabled = rate_limit_strategy != "none"
        self._rate_limiting_backoff_rate = env.MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF_RATE

        self._initial_cpu = psutil.cpu_percent()

    async def connect_async(
        self, 
        cert_path: Optional[str] = None, 
        key_path: Optional[str] = None, 
        worker_socket: Optional[socket.socket] = None,
        worker_server: Optional[asyncio.Server]=None
    ):
        self._backoff_sem = asyncio.Semaphore(
            self._rate_limiting_backoff_rate
        )

        return await super().connect_async(
            cert_path, 
            key_path, 
            worker_socket
        )

    async def connect_client(
        self,
        address: Tuple[str, int],
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        worker_socket: Optional[socket.socket]=None,
        is_ssl: bool=False,
        hostname: str=None,
    ) -> None:
        
        self._hostnames[address] = hostname
        
        if self._semaphore is None:
            self._semaphore = asyncio.Semaphore(self._max_concurrency)

        if self._compressor is None and self._decompressor is None:
            self._compressor = zstandard.ZstdCompressor()
            self._decompressor = zstandard.ZstdDecompressor()
        
       
        if cert_path and key_path:
            self._client_ssl_context = self._create_client_ssl_context(
                cert_path=cert_path,
                key_path=key_path
            ) 

        elif is_ssl:
            self._client_ssl_context = self._create_general_client_ssl_context(
                cert_path=cert_path,
                key_path=key_path
            ) 

        last_error: Union[Exception, None] = None

        for _ in range(self._tcp_connect_retries):

            try:

                self._connections[address] = await asyncio.gather(*[
                    self._connect_client(
                        address,
                        hostname=hostname,
                        worker_socket=worker_socket
                    ) for _ in range(self._max_concurrency)
                ])

                return
            
            except ConnectionRefusedError as connection_error:
                last_error = connection_error

            await asyncio.sleep(1)

        if last_error:
            raise last_error

    def _create_general_client_ssl_context(
        self,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
    ):
        ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        return ctx

        
    async def _connect_client(
        self,
        address: Tuple[str, int],
        hostname: str=None,
        worker_socket: Optional[socket.socket]=None,
    ) -> asyncio.Transport:
        
        self._loop = asyncio.get_event_loop()

        if worker_socket is None:

            http_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            http_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            await self._loop.run_in_executor(None, http_socket.connect, address)

            http_socket.setblocking(False)

        else:
            http_socket = worker_socket
 
        transport, _ = await self._loop.create_connection(
            lambda: MercurySyncTCPClientProtocol(
                self.read
            ),
            sock=http_socket,
            server_hostname=hostname,
            ssl=self._client_ssl_context
        )

        return transport
    
    async def send(
        self, 
        event_name: str,
        data: HTTPRequest, 
        address: Tuple[str, int]
    ):
        async with self._semaphore:

            connections = self._connections.get(address)
            if connections is None:

                connections = await self.connect_client(
                    address,
                    cert_path=self._client_cert_path,
                    key_path=self._client_key_path,
                    is_ssl='https' in data.url
                )

                self._connections[address] = connections

            client_transport = connections.pop()

            result: Union[bytes, None] = None

            try:

                encoded_request = data.prepare_request()
                encrypted_request = self._encryptor.encrypt(encoded_request)
                compressed_request = self._compressor.compress(encrypted_request)

                client_transport.write(compressed_request)

                waiter = self._loop.create_future()
                self._waiters.append(waiter)
                
                result = await waiter

            except Exception:
                self._connections[address].append(
                    await self._connect_client(
                        (
                            self.host,
                            self.port
                        ),
                        hostname=self._hostnames.get(address)
                    )
                )

            self._connections[address].append(client_transport)

            return result
    
    async def send_request(
        self, 
        data: HTTPRequest, 
        address: Tuple[str, int]
    ):
        async with self._semaphore:

            encoded_request = data.prepare_request()

            connections = self._connections.get(address)
            client_transport = connections.pop()

            result: Union[bytes, None] = None

            try:

                client_transport.write(encoded_request)

                waiter = self._loop.create_future()
                self._waiters.append(waiter)
                
                result = await waiter

            except Exception:
                self._connections[address].append(
                    await self._connect_client(
                        (
                            self.host,
                            self.port
                        ),
                        hostname=self._hostnames.get(address)
                    )
                )

            self._connections[address].append(client_transport)

            return result

    def read(
        self, 
        data: bytes, 
        transport: asyncio.Transport
    ) -> None:
        
        if self._is_server:
            self._pending_responses.append(
                asyncio.create_task(
                    self._route_request(
                        data,
                        transport
                    )
                )
            )

        elif bool(self._waiters):

            waiter = self._waiters.pop()
            waiter.set_result(
                HTTPRequest.parse(data)
            )

    async def _route_request(
        self, 
        data: bytes,
        transport: asyncio.Transport
    ):
        if self._use_encryption:
            encrypted_data = self._encryptor.encrypt(data)
            data = self._compressor.compress(encrypted_data)
        
        request_data = data.split(b'\r\n')
        method, path, request_type = request_data[0].decode().split(' ')

        try:

            handler_key = f'{method}_{path}'

            handler = self.events[handler_key]

            query: Union[str, None] = None
            if '?' in path:
                path, query = path.split('?')


            request = Request(
                path,
                method,
                query,
                request_data,
                model=self.parsers.get(handler_key)
            )

            if self._rate_limiting_enabled:

                ip_address, _ = transport.get_extra_info('peername')

                rejected = await self._limiter.limit(
                    ipaddress.ip_address(ip_address),
                    request,
                    limit=handler.limit,
                )

                if rejected and transport.is_closing() is False:

                    async with self._backoff_sem:
                        too_many_requests_response = HTTPMessage(
                            path=request.path,
                            status=429,
                            error='Too Many Requests',
                            protocol=request_type,
                            method=request.method
                        )

                        transport.write(too_many_requests_response.prepare_response())

                        return
                
                elif rejected:

                    async with self._backoff_sem:
                        transport.close()

                        return

            response_info: Tuple[
                Union[
                    Response,
                    BaseModel,
                    str, 
                    None
                ],
                int
            ] = await handler(request)

            (
                response_data, 
                status_code
            ) = response_info

            response_key = f'{handler_key}_{status_code}'
            
            encoded_data: str = ''

            response_parser = self._response_parsers.get(response_key)
            middleware_enabled = self._middleware_enabled.get(path)
            response_headers: Dict[str, str] = handler.response_headers

            if middleware_enabled and response_parser:

                encoded_data = response_parser(response_data.data)
                response_headers.update(
                    response_data.headers
                )

                content_length = len(encoded_data)
                headers = f'content-length: {content_length}'

            elif middleware_enabled:

                encoded_data = response_data.data or ''

                response_headers.update(
                    response_data.headers
                )

                content_length = len(encoded_data)
                headers = f'content-length: {content_length}'

            elif response_parser:

                encoded_data = response_parser(response_data)

                content_length = len(encoded_data)
                headers = f'content-length: {content_length}'


            elif response_data:
                encoded_data = response_data

                content_length = len(response_data)
                headers = f'content-length: {content_length}'

            else:
                headers = 'content-length: 0'

            for key in response_headers:
                headers = f'{headers}\r\n{key}: {response_headers[key]}'

            response_data = f'HTTP/1.1 {status_code} OK\r\n{headers}\r\n\r\n{encoded_data}'.encode()
            
            if self._use_encryption:
                encrypted_data = self._encryptor.encrypt(response_data)
                response_data = self._compressor.compress(encrypted_data)

            transport.write(response_data)

        except KeyError:

            if self._supported_handlers.get(path) is None:
            
                not_found_response = HTTPMessage(
                    path=path,
                    status=404,
                    error='Not Found',
                    protocol=request_type,
                    method=method
                )

                transport.write(not_found_response.prepare_response())

            elif self._supported_handlers[path].get(method) is None:

                method_not_allowed_response = HTTPMessage(
                    path=path,
                    status=405,
                    error='Method Not Allowed',
                    protocol=request_type,
                    method=method
                )

                transport.write(method_not_allowed_response.prepare_response())

        except Exception:

            async with self._backoff_sem:
                if transport.is_closing() is False:

                    server_error_respnse = HTTPMessage(
                        path=path,
                        status=500,
                        error='Internal Error',
                        protocol=request_type,
                        method=method
                    )

                    transport.write(server_error_respnse.prepare_response())

    async def close(self):
        await self._limiter.close()
        return await super().close()




from __future__ import annotations
from collections import defaultdict
from typing import Any, Dict, TypeVar, Generic, List
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.hooks.types.base.hook import Hook
from .event_types import EventType


T = TypeVar('T')


class BaseEvent(Generic[T]):

    __slots__ = (
        'target',
        'event_type',
        'source',
        'as_hook',
        'event_name',
        'event_order',
        'target_is_event',
        'context',
        'events',
        'execution_path',
        'next_args',
        'previous_map',
        'next_map'
    )

    def __init__(self, target: Hook, source: Hook) -> None:
        
        self.target = target
        self.target_is_event = False

        if isinstance(target, BaseEvent):
            self.target_is_event = True

        self.event_type = EventType.EVENT
        self.event_name = source.name
        self.event_order = source.order
        self.event_skip = source.skip

        self.source: T = source

        if target:
            self.target_name = self.target.name
            self.target_shortname = self.target.shortname

        self.as_hook = False
        self.context: SimpleContext = SimpleContext()
        self.events: Dict[str, BaseEvent] = {}
        self.execution_path: List[List[str]] = []
        self.previous_map: List[str] = []
        self.next_map: List[str] = []
        self.next_args: Dict[str, Dict[str,Any]] = defaultdict(dict)

    def __getattribute__(self, name: str) -> Any:
        
        source = None
        event_attrs = [
            'call', 
            'event_type',
            'target', 
            'source',
            'as_hook',
            'event_name',
            'event_order',
            'target_is_event',
            'context',
            'events',
            'execution_path',
            'next_args',
            'previous_map',
            'next_map',
            'copy'
        ]
      
        source = object.__getattribute__(self, 'source')
        
        if source and hasattr(source, name) and name not in event_attrs:
            return getattr(source, name)
        
        return object.__getattribute__(self, name)

    def __setattr__(self, name: str, value: Any) -> None:

        try:

            source = object.__getattribute__(self, 'source')

            event_attrs = [
                'call', 
                'event_type',
                'target', 
                'source',
                'as_hook',
                'event_name',
                'event_order',
                'target_is_event',
                'context',
                'events',
                'execution_path',
                'next_args',
                'previous_map',
                'next_map',
                'copy'
            ]

            if source and hasattr(source, name) and name not in event_attrs:
                return setattr(source, name, value)

        except AttributeError:
            pass

        return super().__setattr__(name, value)

    async def call(self, **kwargs) -> Dict[str, Any]: 
   
        if len(self.next_args[self.event_name]) == 0:
            self.next_args[self.event_name] = kwargs

        if isinstance(self.source, BaseEvent):
            self.source.context = self.context

        results = await self.source.call(**self.next_args[self.event_name])
        

        self.context.update(results)
        self.source.stage_instance.context.update(results)

        if self.source.context:
            self.source.context.update(results)

        next_events = [
            self.events.get(event_name) for event_name in  self.next_map if self.events.get(event_name) is not None
        ]

        for event in next_events:
            event.context = self.context

            self.next_args[event.event_name].update(results)

        return {
            self.event_name: results
        }

    async def execute_pre(self, *hook_args: List[Any]):
        results = None
        for source_name in self.previous_map:
            source: BaseEvent = self.events.get(source_name)
            source.context = self.context
            results = await source.call(*hook_args)

        return results

    async def execute_post(self, results):
        results = None
        for source_name in self.previous_map:
            source: BaseEvent = self.events.get(source_name)
            source.context = self.context
            results = await source.call(results)

        return results

    def copy(self) -> BaseEvent[Hook]:
        base_event = BaseEvent(
            self.target.copy(),
            self.source.copy()
        )

        base_event.execution_path = self.execution_path
        base_event.previous_map = self.previous_map
        base_event.next_map = self.next_map
        base_event.next_args = self.next_args

        return base_event



from typing import Union
from pydantic import (
    BaseModel, 
    StrictInt, 
    StrictFloat, 
    validator
)


class ParameterRange(BaseModel):
    minimum_range: Union[StrictInt, StrictFloat]
    maximum_range: Union[StrictInt, StrictFloat]

    class Config:
        arbitrary_types_allowed = True

    @validator('minimum_range', 'maximum_range')
    def validate_param_range(cls, val):
        assert val > 0, "Order and weight values must be greater than zero!"

        return val



from ssl import SSLContext
from typing import List, Optional
from pydantic import BaseModel, StrictStr, StrictInt
from hedra.data.connectors.common.connector_type import ConnectorType


class CassandraConnectorConfig(BaseModel):
    hosts: List[StrictStr] = ['127.0.0.1']
    port: StrictInt=9042
    username: Optional[StrictStr]=None
    password: Optional[StrictStr]=None
    keyspace: StrictStr='hedra'
    table_name: StrictStr
    replication_strategy: StrictStr='SimpleStrategy'
    replication: StrictInt=3
    ssl: Optional[SSLContext]=None
    connector_type: ConnectorType=ConnectorType.Cassandra

    class Config:
        arbitrary_types_allowed=True


from pydantic import BaseModel, StrictStr, StrictInt
from typing import Dict, Any, Optional, Type

class FieldOption:
    field_type: StrictStr
    options: Dict[StrictStr, Any]


class CassandraLoadValidator(BaseModel):
    fields: Optional[Dict[StrictStr, FieldOption]]
    filters: Optional[Dict[StrictStr, Any]]
    table: Type[Any]
    limit: Optional[StrictInt]

    class Config:
        arbitrary_types_allowed=True


from typing import Any
from .flag_type import FlagTypes


class Flag:

    def __init__(
            self,
            feature_name: str,
            feature: Any,
            flag_type: FlagTypes,
            exception: Exception,
            enabled: bool=False
        ) -> None:
        self.name = feature_name
        self.feature = feature
        self.type = flag_type
        self.exception = exception
        self.enabled = enabled


from typing import Any
from hedra.versioning.flags.types.base.feature import Flag
from hedra.versioning.flags.types.base.flag_type import FlagTypes
from hedra.versioning.flags.exceptions.latest_not_enabled import LatestNotEnabledException


class UnstableFeature(Flag):

    def __init__(
            self,
            feature_name: str,
            feature: Any,
        ) -> None:
        super().__init__(
            feature_name,
            feature,
            FlagTypes.UNSTABLE_FEATURE,
            LatestNotEnabledException,
        )
        


from __future__ import annotations
import asyncio
from typing import List
from hedra.core.graphs.transitions.common.base_edge import BaseEdge
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.complete.complete import Complete
from hedra.core.graphs.stages.types.stage_states import StageStates


class CompleteEdge(BaseEdge[Complete]):

    def __init__(self, source: Complete, destination: BaseEdge[Stage]) -> None:
        super(
            CompleteEdge,
            self
        ).__init__(
            source,
            destination
        )

    async def transition(self):

        await self.source.run()

        self.source.state = StageStates.COMPLETE

        self.visited.append(self.source.name)

        return None, None

    def _update(self, destination: Stage):
        self.next_history.update({
            (self.source.name, destination.name): {}
        })

    def split(self, edges: List[CompleteEdge]) -> None:
        pass
        




import asyncio
import math
import statistics
from hedra.distributed.models.http import Limit
from .base_limiter import BaseLimiter

class AdaptiveRateLimiter(BaseLimiter):

    __slots__ = (
        "max_rate",
        "min_rate",
        "time_period",
        "history",
        "rate_history",
        "moments",
        "waiting",
        "last_request_time",
        "current_rate",
        "_rate_per_sec",
        "_level",
        "_waiters",
        "_loop",
        "_current_time",
        "_previous_count",
        "_last_slope",
        "_current_slope"
    )

    def __init__(
        self, 
        limit: Limit
    ):
        super().__init__(
            limit.max_requests,
            limit.period
        )

        min_requests = limit.min_requests
        if min_requests is None:
            min_requests = math.ceil(self.max_rate * 0.1)

        self.initial_rate = math.ceil(
            (self.max_rate - min_requests)/2
        )

        self.min_rate = min_requests

        self.history = []
        self.rate_history = []
        self.moments = []
        self.waiting = []


        self._loop = asyncio.get_event_loop()

        self._current_time = self._loop.time()
        self._previous_count = limit.max_requests

        self.last_request_time = self._loop.time()
        self.current_rate = self.initial_rate

    def get_next_rate(self):

        current_time = self._loop.time()

        elapsed_time = current_time - self.last_request_time
        self.history.append(elapsed_time)

        if len(self.history) > self.time_period:
            self.history.pop(0)

        average_time = statistics.mean(self.history)

        if average_time > 1 / self.current_rate:
            self.current_rate = max(self.min_rate, self.current_rate / 2)
        else:
            self.current_rate = min(self.max_rate, self.current_rate * 2)

        self.last_request_time = current_time

        return self.current_rate

    def has_capacity(self, amount: float = 1) -> bool:

        expected_rate = self.get_next_rate()

        if (self._loop.time() - self._current_time) > self.time_period:
            self._current_time = math.floor(
                self._loop.time()/self.time_period
            ) * self.time_period

            self._previous_count = self._level
            self._level = 0

        self._rate_per_sec = (
            self._previous_count * (
                self.time_period - (self._loop.time() - self._current_time)
            )/self.time_period
        ) + (self._level + amount)

        if self._rate_per_sec < expected_rate:

            for fut in self._waiters.values():
                if not fut.done():
                    fut.set_result(True)
                    break

        return self._rate_per_sec <= expected_rate



from enum import Enum

class StreamStateMap(Enum):
    SEND_HEADERS = 0
    SEND_PUSH_PROMISE = 1
    SEND_RST_STREAM = 2
    SEND_DATA = 3
    SEND_WINDOW_UPDATE = 4
    SEND_END_STREAM = 5
    RECV_HEADERS = 6
    RECV_PUSH_PROMISE = 7
    RECV_RST_STREAM = 8
    RECV_DATA = 9
    RECV_WINDOW_UPDATE = 10
    RECV_END_STREAM = 11
    RECV_CONTINUATION = 12  # Added in 2.0.0
    SEND_INFORMATIONAL_HEADERS = 13  # Added in 2.2.0
    RECV_INFORMATIONAL_HEADERS = 14  # Added in 2.2.0
    SEND_ALTERNATIVE_SERVICE = 15  # Added in 2.3.0
    RECV_ALTERNATIVE_SERVICE = 16  # Added in 2.3.0
    UPGRADE_CLIENT = 17  # Added 2.3.0
    UPGRADE_SERVER = 18  # Added 2.3.0



from pydantic import BaseModel
from hedra.data.connectors.common.connector_type import ConnectorType


class GoogleCloudStorageConnectorConfig(BaseModel):
    service_account_json_path: str
    bucket_namespace: str
    bucket_name: str
    connector_type: ConnectorType=ConnectorType.GCS


from typing import Any
from hedra.core.engines.types.grpc import GRPCResult
from .http2_processed_result import HTTP2ProcessedResult


class GRPCProcessedResult(HTTP2ProcessedResult):

    def __init__(
        self, 
        stage: str, 
        result: GRPCResult
    ) -> None:
        super(
            GRPCProcessedResult,
            self
        ).__init__(
            stage,
            result
        )


from typing import Optional
from hedra.data.connectors.common.connector_type import ConnectorType
from pydantic import (
    BaseModel,
    StrictStr
)


class MongoDBConnectorConfig(BaseModel):
    host: StrictStr='localhost:27017'
    username: Optional[StrictStr]
    password: Optional[StrictStr]
    database: StrictStr
    collection: StrictStr
    connector_type: ConnectorType=ConnectorType.MongoDB


from typing import Dict, Generic, Union
from typing_extensions import TypeVarTuple, Unpack
from hedra.core.engines.client.config import Config
from hedra.core.experiments.mutations.types.base.mutation import Mutation
from .store import ActionsStore

T = TypeVarTuple('T')


class PluginsStore(Generic[Unpack[T]]):

    def __init__(self, metadata_string: str):
        self._plugins: Dict[str, Union[Unpack[T]]] = {}
        self._config: Config = None
        self.actions = ActionsStore(metadata_string)
        self.next_name: str = None
        self.intercept: bool = False
        self.metadata_string: str = metadata_string
        self.clients = {}
        self.mutations: Dict[str, Mutation] = {}

    def __getitem__(self, plugin_name: str) -> Union[Unpack[T]]:

        custom_plugin: Union[Unpack[T]] = self._plugins.get(plugin_name)

        if custom_plugin.initialized is False:
            custom_plugin.name = plugin_name
            custom_plugin.actions = self.actions
            custom_plugin.initialized = True
            custom_plugin.mutations.update(self.mutations)

        custom_plugin.metadata_string = self.metadata_string
        custom_plugin.next_name = self.next_name
        custom_plugin.intercept = self.intercept

        self._plugins[plugin_name] = custom_plugin
        self.clients[plugin_name] = custom_plugin

        return custom_plugin

    def __setitem__(self, plugin_name: str, plugin: Union[Unpack[T]]) -> None:
        self._plugins[plugin_name] = plugin